import pytest
import torch

from aspire.flows.torch.flows import ZukoFlow
from aspire.transforms import FlowTransform
from aspire.utils import AspireFile


@pytest.mark.parametrize("dtype", [torch.float32, torch.float64])
def test_zuko_flow(dtype):
    dims = 3
    parameters = [f"x_{i}" for i in range(dims)]

    data_transform = FlowTransform(
        parameters=parameters, xp=torch, dtype=dtype
    )

    # Create an instance of ZukoFlow
    flow = ZukoFlow(
        dims=dims,
        flow_class="MAF",
        seed=42,
        device="cpu",
        data_transform=data_transform,
    )

    x = torch.randn(100, dims, device=flow.device)

    flow.fit_data_transform(x)

    # Check if the flow is initialized correctly
    assert flow.dims == dims

    # Check if the flow is an instance of ZukoFlow
    assert isinstance(flow, ZukoFlow)

    # Check if the flow has a valid flow attribute
    assert flow.flow is not None

    x = torch.tensor([0.1, 0.2, 0.3], device=flow.device)

    log_prob = flow.log_prob(x)

    assert log_prob.shape == (1,)


def test_zuko_flow_save_and_load(tmp_path):
    flow = ZukoFlow(
        dims=2,
        flow_class="MAF",
        seed=42,
        device="cpu",
    )

    x = torch.randn(100, 2, device=flow.device)

    with AspireFile(tmp_path / "result.h5", "w") as f:
        flow.save(f, "flow")

    with AspireFile(tmp_path / "result.h5", "r") as f:
        loaded_flow = ZukoFlow.load(f, "flow")

    # Check if the loaded flow is equivalent to the original flow
    assert loaded_flow.dims == flow.dims
    assert loaded_flow.flow is not None

    log_prob_original = flow.log_prob(x)
    log_prob_loaded = loaded_flow.log_prob(x)

    assert torch.allclose(log_prob_original, log_prob_loaded)



import math
from collections import namedtuple

import pytest

from aspire.samples import Samples

SamplerConfig = namedtuple(
    "SamplerConfig",
    ["sampler", "sampler_kwargs"],
)


@pytest.fixture
def dims():
    return 2


@pytest.fixture
def parameters(dims):
    return [f"x_{i}" for i in range(dims)]


@pytest.fixture
def prior_bounds(parameters):
    return {p: [-10, 10] for p in parameters}


@pytest.fixture
def n_samples():
    return 500


@pytest.fixture
def likelihood_mean():
    return 2.0


@pytest.fixture
def likelihood_std():
    return 1.0


@pytest.fixture
def initial_samples(dims, rng, likelihood_mean, likelihood_std, n_samples):
    return rng.normal(likelihood_mean, likelihood_std, size=(n_samples, dims))


@pytest.fixture
def samples(initial_samples, xp):
    assert initial_samples.ndim == 2
    return Samples(initial_samples, xp=xp)


@pytest.fixture(params=[True, False])
def bounded_to_unbounded(request):
    return request.param


@pytest.fixture(params=["zuko", "flowjax"])
def flow_backend(request):
    return request.param


@pytest.fixture(params=["jax", "numpy", "torch"])
def samples_backend(request):
    return request.param


@pytest.fixture
def xp(samples_backend):
    if samples_backend == "jax":
        import jax.numpy as xp
    elif samples_backend == "torch":
        import array_api_compat.torch as xp
    elif samples_backend == "numpy":
        import array_api_compat.numpy as xp
    else:
        raise ValueError(f"Unsupported backend: {samples_backend}")
    return xp


@pytest.fixture
def log_likelihood(likelihood_mean, likelihood_std, xp, dims):
    def _log_likelihood(samples):
        assert samples.x.shape[-1] == dims
        x = xp.asarray(samples.x)
        constant = xp.log(
            xp.asarray(1 / (likelihood_std * math.sqrt(2 * math.pi)))
        )
        return xp.sum(
            constant - (0.5 * ((x - likelihood_mean) / likelihood_std) ** 2),
            axis=-1,
        )

    return _log_likelihood


@pytest.fixture
def log_prior(dims, xp):
    def _log_prior(samples):
        assert samples.x.shape[-1] == dims
        x = xp.asarray(samples.x)
        constant = dims * xp.log(xp.asarray(1 / 10))
        val = xp.where((x >= -10) & (x <= 10), constant, -xp.inf)
        return xp.sum(val, axis=-1)

    return _log_prior


@pytest.fixture(
    params=[
        "importance",
        "minipcn_smc",
        "emcee_smc",
        "smc",
        "emcee",
        "minipcn",
        "blackjax_smc_rwmh",
        "blackjax_smc_nuts",
    ]
)
def sampler_config(request):
    if request.param == "importance":
        return SamplerConfig(sampler="importance", sampler_kwargs={})
    elif request.param == "emcee":
        return SamplerConfig(
            sampler="emcee",
            sampler_kwargs={
                "nsteps": 100,
            },
        )
    elif request.param == "minipcn":
        return SamplerConfig(
            sampler="minipcn",
            sampler_kwargs={
                "n_steps": 500,
            },
        )
    elif request.param in ["smc", "minipcn_smc"]:
        return SamplerConfig(
            sampler="minipcn_smc",
            sampler_kwargs={
                "adaptive": True,
                "sampler_kwargs": {
                    "n_steps": 10,
                },
            },
        )
    elif request.param == "emcee_smc":
        return SamplerConfig(
            sampler="emcee_smc",
            sampler_kwargs={
                "adaptive": True,
                "sampler_kwargs": {
                    "nsteps": 10,
                    "progress": False,
                },
            },
        )
    elif request.param == "blackjax_smc_rwmh":
        return SamplerConfig(
            sampler="blackjax_smc",
            sampler_kwargs={
                "adaptive": True,
                "sampler_kwargs": {
                    "algorithm": "rwmh",
                    "sigma": 0.1,
                    "n_steps": 10,
                },
            },
        )
    elif request.param == "blackjax_smc_nuts":
        return SamplerConfig(
            sampler="blackjax_smc",
            sampler_kwargs={
                "adaptive": True,
                "sampler_kwargs": {
                    "algorithm": "nuts",
                    "step_size": 0.1,
                    "n_steps": 10,
                },
            },
        )
    else:
        raise ValueError(f"Unsupported sampler: {request.param}")



from ..samples import Samples
from ..utils import track_calls
from .base import Sampler


class ImportanceSampler(Sampler):
    @track_calls
    def sample(self, n_samples: int) -> Samples:
        x, log_q = self.prior_flow.sample_and_log_prob(n_samples)
        samples = Samples(
            x,
            log_q=log_q,
            xp=self.xp,
            parameters=self.parameters,
            dtype=self.dtype,
        )
        samples.log_prior = samples.array_to_namespace(self.log_prior(samples))
        samples.log_likelihood = samples.array_to_namespace(
            self.log_likelihood(samples)
        )
        samples.compute_weights()
        return samples



"""Sphinx configuration for the aspire documentation."""

from __future__ import annotations

import os
import sys

# Ensure the package can be imported when building the docs locally
ROOT = os.path.abspath("..")
if ROOT not in sys.path:
    sys.path.insert(0, ROOT)

import aspire  # noqa: E402

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "aspire"
copyright = "2025, Michael J. Williams"
author = "Michael J. Williams"
version = aspire.__version__
release = aspire.__version__

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "autoapi.extension",
]

autodoc_typehints = "description"
autodoc_member_order = "bysource"
autosummary_generate = True
napoleon_google_docstring = False
napoleon_numpy_docstring = True
napoleon_preprocess_types = True

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# -- Configure autoapi -------------------------------------------------------

autoapi_type = "python"
autoapi_dirs = ["../src/aspire/"]
autoapi_add_toctree_entry = True
autoapi_options = [
    "members",
    "imported-members",
    "show-inheritance",
    "show-module-summary",
    "undoc-members",
]

# -- Options for HTML output -------------------------------------------------

html_theme = "sphinx_book_theme"
html_static_path = ["_static"]
html_title = "aspire"
html_theme_options = {
    "path_to_docs": "docs",
    "repository_url": "https://github.com/mj-will/aspire",
    "repository_branch": "main",
    "use_edit_page_button": True,
    "use_issues_button": True,
    "use_repository_button": True,
    "use_download_button": True,
}



import logging
from typing import Any, Callable

import array_api_compat.numpy as np

from ...flows.base import Flow
from ...history import SMCHistory
from ...samples import SMCSamples
from ...utils import (
    asarray,
    effective_sample_size,
    track_calls,
    update_at_indices,
)
from ..mcmc import MCMCSampler

logger = logging.getLogger(__name__)


class SMCSampler(MCMCSampler):
    """Base class for Sequential Monte Carlo samplers."""

    def __init__(
        self,
        log_likelihood: Callable,
        log_prior: Callable,
        dims: int,
        prior_flow: Flow,
        xp: Callable,
        dtype: Any | str | None = None,
        parameters: list[str] | None = None,
        rng: np.random.Generator | None = None,
        preconditioning_transform: Callable | None = None,
    ):
        super().__init__(
            log_likelihood=log_likelihood,
            log_prior=log_prior,
            dims=dims,
            prior_flow=prior_flow,
            xp=xp,
            dtype=dtype,
            parameters=parameters,
            preconditioning_transform=preconditioning_transform,
        )
        self.rng = rng or np.random.default_rng()
        self._adapative_target_efficiency = False

    @property
    def target_efficiency(self):
        return self._target_efficiency

    @target_efficiency.setter
    def target_efficiency(self, value: float | tuple):
        """Set the target efficiency.

        Parameters
        ----------
        value : float or tuple
            If a float, the target efficiency to use for all iterations.
            If a tuple of two floats, the target efficiency will adapt from
            the first value to the second value over the course of the SMC
            iterations. See `target_efficiency_rate` for details.
        """
        if isinstance(value, float):
            if not (0 < value < 1):
                raise ValueError("target_efficiency must be in (0, 1)")
            self._target_efficiency = value
            self._adapative_target_efficiency = False
        elif len(value) != 2:
            raise ValueError(
                "target_efficiency must be a float or tuple of two floats"
            )
        else:
            value = tuple(map(float, value))
            if not (0 < value[0] < value[1] < 1):
                raise ValueError(
                    "target_efficiency tuple must be in (0, 1) and increasing"
                )
            self._target_efficiency = value
            self._adapative_target_efficiency = True

    def current_target_efficiency(self, beta: float) -> float:
        """Get the current target efficiency based on beta."""
        if self._adapative_target_efficiency:
            return self._target_efficiency[0] + (
                self._target_efficiency[1] - self._target_efficiency[0]
            ) * (beta**self.target_efficiency_rate)
        else:
            return self._target_efficiency

    def determine_beta(
        self,
        samples: SMCSamples,
        beta: float,
        beta_step: float,
        min_step: float,
    ) -> tuple[float, float]:
        """Determine the next beta value.

        Parameters
        ----------
        samples : SMCSamples
            The current samples.
        beta : float
            The current beta value.
        beta_step : float
            The fixed beta step size if not adaptive.
        min_step : float
            The minimum beta step size.

        Returns
        -------
        beta : float
            The new beta value.
        min_step : float
            The new minimum step size if adaptive_min_step is True.
        """
        if not self.adaptive:
            beta += beta_step
            if beta >= 1.0:
                beta = 1.0
        else:
            beta_prev = beta
            beta_min = beta_prev
            beta_max = 1.0
            tol = 1e-5
            eff_beta_max = effective_sample_size(
                samples.log_weights(beta_max)
            ) / len(samples)
            if eff_beta_max >= self.current_target_efficiency(beta_prev):
                beta_min = 1.0
            target_eff = self.current_target_efficiency(beta_prev)
            while beta_max - beta_min > tol:
                beta_try = 0.5 * (beta_max + beta_min)
                eff = effective_sample_size(
                    samples.log_weights(beta_try)
                ) / len(samples)
                if eff >= target_eff:
                    beta_min = beta_try
                else:
                    beta_max = beta_try
            beta_star = beta_min

            if self.adaptive_min_step:
                min_step = min_step * (1 - beta_prev) / (1 - beta_star)
            beta = max(beta_star, beta_prev + min_step)
            beta = min(beta, 1.0)
        return beta, min_step

    @track_calls
    def sample(
        self,
        n_samples: int,
        n_steps: int | None = None,
        adaptive: bool = True,
        min_step: float | None = None,
        max_n_steps: int | None = None,
        target_efficiency: float = 0.5,
        target_efficiency_rate: float = 1.0,
        n_final_samples: int | None = None,
    ) -> SMCSamples:
        samples = self.draw_initial_samples(n_samples)
        samples = SMCSamples.from_samples(
            samples, xp=self.xp, beta=0.0, dtype=self.dtype
        )
        self.fit_preconditioning_transform(samples.x)

        if self.xp.isnan(samples.log_q).any():
            raise ValueError("Log proposal contains NaN values")
        if self.xp.isnan(samples.log_prior).any():
            raise ValueError("Log prior contains NaN values")
        if self.xp.isnan(samples.log_likelihood).any():
            raise ValueError("Log likelihood contains NaN values")

        logger.debug(f"Initial sample summary: {samples}")

        # Remove the n_final_steps from sampler_kwargs if present
        self.sampler_kwargs = self.sampler_kwargs or {}
        n_final_steps = self.sampler_kwargs.pop("n_final_steps", None)

        self.history = SMCHistory()

        self.target_efficiency = target_efficiency
        self.target_efficiency_rate = target_efficiency_rate

        if n_steps is not None:
            beta_step = 1 / n_steps
        elif not adaptive:
            raise ValueError("Either n_steps or adaptive=True must be set")
        else:
            beta_step = np.nan
        self.adaptive = adaptive
        beta = 0.0

        if min_step is None:
            if max_n_steps is None:
                min_step = 0.0
                self.adaptive_min_step = False
            else:
                min_step = 1 / max_n_steps
                self.adaptive_min_step = True
        else:
            self.adaptive_min_step = False

        iterations = 0
        while True:
            iterations += 1

            beta, min_step = self.determine_beta(
                samples,
                beta,
                beta_step,
                min_step,
            )
            self.history.eff_target.append(
                self.current_target_efficiency(beta)
            )

            logger.info(f"it {iterations} - beta: {beta}")
            self.history.beta.append(beta)

            ess = effective_sample_size(samples.log_weights(beta))
            eff = ess / len(samples)
            if eff < 0.1:
                logger.warning(
                    f"it {iterations} - Low sample efficiency: {eff:.2f}"
                )
            self.history.ess.append(ess)
            logger.info(
                f"it {iterations} - ESS: {ess:.1f} ({eff:.2f} efficiency)"
            )
            self.history.ess_target.append(
                effective_sample_size(samples.log_weights(1.0))
            )

            log_evidence_ratio = samples.log_evidence_ratio(beta)
            log_evidence_ratio_var = samples.log_evidence_ratio_variance(beta)
            self.history.log_norm_ratio.append(log_evidence_ratio)
            self.history.log_norm_ratio_var.append(log_evidence_ratio_var)
            logger.info(
                f"it {iterations} - Log evidence ratio: {log_evidence_ratio:.2f} +/- {np.sqrt(log_evidence_ratio_var):.2f}"
            )

            samples = samples.resample(beta, rng=self.rng)

            samples = self.mutate(samples, beta)
            if beta == 1.0 or (
                max_n_steps is not None and iterations >= max_n_steps
            ):
                break

        # If n_final_samples is not None, perform additional mutations steps
        if n_final_samples is not None:
            logger.info(f"Generating {n_final_samples} final samples")
            final_samples = samples.resample(
                1.0, n_samples=n_final_samples, rng=self.rng
            )
            samples = self.mutate(final_samples, 1.0, n_steps=n_final_steps)

        samples.log_evidence = samples.xp.sum(
            asarray(self.history.log_norm_ratio, self.xp)
        )
        samples.log_evidence_error = samples.xp.sqrt(
            samples.xp.sum(asarray(self.history.log_norm_ratio_var, self.xp))
        )

        final_samples = samples.to_standard_samples()
        logger.info(
            f"Log evidence: {final_samples.log_evidence:.2f} +/- {final_samples.log_evidence_error:.2f}"
        )
        return final_samples

    def mutate(self, particles):
        raise NotImplementedError

    def log_prob(self, z, beta=None):
        x, log_abs_det_jacobian = self.preconditioning_transform.inverse(z)
        samples = SMCSamples(x, xp=self.xp, beta=beta, dtype=self.dtype)
        log_q = self.prior_flow.log_prob(samples.x)
        samples.log_q = samples.array_to_namespace(log_q)
        samples.log_prior = self.log_prior(samples)
        samples.log_likelihood = self.log_likelihood(samples)
        log_prob = samples.log_p_t(
            beta=beta
        ).flatten() + samples.array_to_namespace(log_abs_det_jacobian)

        log_prob = update_at_indices(
            log_prob, self.xp.isnan(log_prob), -self.xp.inf
        )
        return log_prob


class NumpySMCSampler(SMCSampler):
    def __init__(
        self,
        log_likelihood,
        log_prior,
        dims,
        prior_flow,
        xp,
        dtype=None,
        parameters=None,
        preconditioning_transform=None,
    ):
        if preconditioning_transform is not None:
            preconditioning_transform = preconditioning_transform.new_instance(
                xp=np
            )
        super().__init__(
            log_likelihood,
            log_prior,
            dims,
            prior_flow=prior_flow,
            xp=xp,
            dtype=dtype,
            parameters=parameters,
            preconditioning_transform=preconditioning_transform,
        )



import copy
import logging

import numpy as np

from ...samples import SMCSamples
from ...utils import track_calls
from .base import NumpySMCSampler

logger = logging.getLogger(__name__)


class EmceeSMC(NumpySMCSampler):
    @track_calls
    def sample(
        self,
        n_samples: int,
        n_steps: int = None,
        adaptive: bool = True,
        target_efficiency: float = 0.5,
        target_efficiency_rate: float = 1.0,
        sampler_kwargs: dict | None = None,
        n_final_samples: int | None = None,
    ):
        self.sampler_kwargs = sampler_kwargs or {}
        self.sampler_kwargs.setdefault("nsteps", 5 * self.dims)
        self.sampler_kwargs.setdefault("progress", True)
        self.emcee_moves = self.sampler_kwargs.pop("moves", None)
        return super().sample(
            n_samples,
            n_steps=n_steps,
            adaptive=adaptive,
            target_efficiency=target_efficiency,
            target_efficiency_rate=target_efficiency_rate,
            n_final_samples=n_final_samples,
        )

    def mutate(self, particles, beta, n_steps=None):
        import emcee

        logger.info("Mutating particles")
        sampler = emcee.EnsembleSampler(
            len(particles.x),
            self.dims,
            self.log_prob,
            args=(beta,),
            vectorize=True,
            moves=self.emcee_moves,
        )
        z = self.fit_preconditioning_transform(particles.x)
        kwargs = copy.deepcopy(self.sampler_kwargs)
        if n_steps is not None:
            kwargs["nsteps"] = n_steps
        sampler.run_mcmc(z, **kwargs)
        self.history.mcmc_acceptance.append(
            np.mean(sampler.acceptance_fraction)
        )
        self.history.mcmc_autocorr.append(
            sampler.get_autocorr_time(
                quiet=True, discard=int(0.2 * self.sampler_kwargs["nsteps"])
            )
        )
        z = sampler.get_chain(flat=False)[-1, ...]
        x = self.preconditioning_transform.inverse(z)[0]
        samples = SMCSamples(x, xp=self.xp, beta=beta, dtype=self.dtype)
        samples.log_q = samples.array_to_namespace(
            self.prior_flow.log_prob(samples.x)
        )
        samples.log_prior = samples.array_to_namespace(self.log_prior(samples))
        samples.log_likelihood = samples.array_to_namespace(
            self.log_likelihood(samples)
        )
        if samples.xp.isnan(samples.log_q).any():
            raise ValueError("Log proposal contains NaN values")
        return samples



import jax
import jax.numpy as jnp
import pytest

from aspire.flows.jax.flows import FlowJax
from aspire.transforms import FlowTransform
from aspire.utils import AspireFile


@pytest.mark.parametrize("dtype", [jnp.float32, jnp.float64])
def test_flowjax_flow(dtype):
    dims = 3
    parameters = [f"x_{i}" for i in range(dims)]

    data_transform = FlowTransform(parameters=parameters, xp=jnp, dtype=dtype)
    key = jax.random.key(42)
    key, flow_key = jax.random.split(key)

    # Create an instance of FlowJax
    flow = FlowJax(
        dims=dims,
        key=flow_key,
        device="cpu",
        data_transform=data_transform,
        dtype=dtype,
    )

    key, samples_key = jax.random.split(key)
    x = jax.random.normal(samples_key, (100, dims))

    flow.fit_data_transform(x)

    # Check if the flow is initialized correctly
    assert flow.dims == dims

    # Check if the flow is an instance of ZukoFlow
    assert isinstance(flow, FlowJax)

    x = jnp.array([0.1, 0.2, 0.3])

    log_prob = flow.log_prob(x)

    assert log_prob.shape == (1,)

    key, sample_key = jax.random.split(key)
    x = flow.sample(1)
    assert x.shape == (1, dims)


def test_flowjax_save_and_load(tmp_path):
    key = jax.random.key(42)
    key, flow_key = jax.random.split(key)
    flow = FlowJax(
        dims=2,
        key=flow_key,
        device="cpu",
        data_transform=FlowTransform(
            parameters=[f"x_{i}" for i in range(2)], xp=jnp
        ),
    )

    key, samples_key = jax.random.split(key)
    x = jax.random.normal(samples_key, (100, 2))

    flow.fit_data_transform(x)

    log_prob = flow.log_prob(x)

    key_data = jax.random.key_data(flow_key)
    re_key = jax.random.wrap_key_data(key_data)
    assert flow_key == re_key

    with AspireFile(tmp_path / "result.h5", "w") as f:
        flow.save(f, "flow")

    with AspireFile(tmp_path / "result.h5", "r") as f:
        loaded_flow = FlowJax.load(f, "flow")

    # Check if the loaded flow is equivalent to the original flow
    assert loaded_flow.dims == flow.dims

    log_prob_loaded = loaded_flow.log_prob(x)
    assert jnp.allclose(log_prob, log_prob_loaded)



import numpy as np
import pytest
from array_api_compat import is_torch_namespace


@pytest.fixture(autouse=True, scope="session")
def enable_jax_float64():
    try:
        import jax

        jax.config.update("jax_enable_x64", True)
    except ImportError:
        pass


@pytest.fixture
def rng():
    return np.random.default_rng()


@pytest.fixture(params=["jax", "torch", "numpy"])
def xp(request):
    if request.param == "jax":
        import jax.numpy as xp
    elif request.param == "torch":
        import array_api_compat.torch as xp
    elif request.param == "numpy":
        import array_api_compat.numpy as xp
    else:
        raise ValueError(f"Unsupported backend: {request.param}")
    return xp


@pytest.fixture(params=["float32", "float64", None])
def dtype(request, xp):
    if request.param is None:
        return None
    if is_torch_namespace(xp):
        import torch

        if request.param == "float32":
            return torch.float32
        elif request.param == "float64":
            return torch.float64
        else:
            raise ValueError(f"Unsupported dtype: {request.param}")
    return xp.dtype(request.param)



from typing import Callable

import flowjax.bijections
import flowjax.distributions
import flowjax.flows
import jax
import jax.numpy as jnp
import jax.random as jrandom


def get_flow_function_class(name: str) -> Callable:
    try:
        return getattr(flowjax.flows, name)
    except AttributeError:
        raise ValueError(f"Unknown flow function: {name}")


def get_bijection_class(name: str) -> Callable:
    try:
        return getattr(flowjax.bijections, name)
    except AttributeError:
        raise ValueError(f"Unknown bijection: {name}")


def get_flow(
    *,
    key: jax.Array,
    dims: int,
    flow_type: str | Callable = "masked_autoregressive_flow",
    bijection_type: str | flowjax.bijections.AbstractBijection | None = None,
    bijection_kwargs: dict | None = None,
    dtype=None,
    **kwargs,
) -> flowjax.distributions.Transformed:
    dtype = dtype or jnp.float32

    if isinstance(flow_type, str):
        flow_type = get_flow_function_class(flow_type)

    if isinstance(bijection_type, str):
        bijection_type = get_bijection_class(bijection_type)
    if bijection_type is not None:
        transformer = bijection_type(**bijection_kwargs)
    else:
        transformer = None

    if bijection_kwargs is None:
        bijection_kwargs = {}

    base_dist = flowjax.distributions.Normal(jnp.zeros(dims, dtype=dtype))
    key, subkey = jrandom.split(key)
    return flow_type(
        subkey,
        base_dist=base_dist,
        transformer=transformer,
        **kwargs,
    )



"""Example using sequential posterior inference with SMC.

This examples is slightly contrived, using a mixture of two Gaussians in 4D
as the target distribution. The goal is to demonstrate the ability of SMC to
explore multi-modal distributions, even when the initial samples deviate
significantly from the true modes.

In practice, one would ideally use more informative initial samples.
"""

from pathlib import Path

import numpy as np

from aspire import Aspire
from aspire.plot import plot_comparison
from aspire.samples import Samples
from aspire.utils import configure_logger

# RNG for generating initial samples
rng = np.random.default_rng(42)

# Output directory
outdir = Path("outdir") / "smc_example"
outdir.mkdir(parents=True, exist_ok=True)

# Configure logger to show INFO level messages
configure_logger()

# Number of dimensions
dims = 4

# Means and covariances of the two Gaussian components
mu1 = 2 * np.ones(dims)
mu2 = -2 * np.ones(dims)
cov1 = 0.5 * np.eye(dims)
cov2 = np.eye(dims)


def log_likelihood(samples):
    """Log-likelihood of a mixture of two Gaussians"""
    x = samples.x
    comp1 = (
        -0.5 * ((x - mu1) @ np.linalg.inv(cov1) * (x - mu1)).sum(axis=-1)
        - 0.5 * dims * np.log(2 * np.pi)
        - 0.5 * np.linalg.slogdet(cov1)[1]
    )
    comp2 = (
        -0.5 * ((x - mu2) @ np.linalg.inv(cov2) * (x - mu2)).sum(axis=-1)
        - 0.5 * dims * np.log(2 * np.pi)
        - 0.5 * np.linalg.slogdet(cov2)[1]
    )
    return np.logaddexp(comp1, comp2)  # Log-sum-exp for numerical stability


def log_prior(samples):
    """Standard normal prior"""
    return -0.5 * (samples.x**2).sum(axis=-1) - dims * 0.5 * np.log(2 * np.pi)


# Generate prior samples for comparison, these are not used in SMC
prior_samples = Samples(rng.normal(0, 1, size=(5000, dims)))

# We draw initial samples from two Gaussians centered away from the true modes
# to demonstrate the ability of SMC to explore the posterior
offset_1 = rng.uniform(-3, 3, size=(dims,))
offset_2 = rng.uniform(-3, 3, size=(dims,))
initial_samples = np.concatenate(
    [
        rng.normal(mu1 - offset_1, 1, size=(2500, dims)),
        rng.normal(mu2 - offset_2, 1, size=(2500, dims)),
    ],
    axis=0,
)
initial_samples = Samples(initial_samples)

# Initialize Aspire with the log-likelihood and log-prior
aspire = Aspire(
    log_likelihood=log_likelihood,
    log_prior=log_prior,
    dims=dims,
    flow_class="NSF",  # Use Neural Spline Flow from zuko (default backend)
)

# Fit the normalizing flow to the initial samples
fit_history = aspire.fit(initial_samples, n_epochs=30)
# Plot loss
fit_history.plot_loss().savefig(outdir / "loss.png")

# Sample from the posterior using SMC
samples, history = aspire.sample_posterior(
    sampler="smc",  # Sequential Monte Carlo, this uses the default minipcn sampler
    n_samples=500,  # Number of particles in SMC
    n_final_samples=5000,  # Number of samples to draw from the final distribution
    sampler_kwargs=dict(  # Keyword arguments for the specific sampler
        n_steps=20,  # MCMC steps per SMC iteration
    ),
    return_history=True,  # To return the SMC history (e.g., ESS, betas)
)
# Plot SMC diagnostics
history.plot().savefig(outdir / "smc_diagnostics.png")

# Plot corner plot of the samples
# Include initial samples and prior samples for comparison
plot_comparison(
    initial_samples,
    prior_samples,
    samples,
    labels=["Initial Samples", "Prior Samples", "SMC Samples"],
).savefig(outdir / "posterior.png")



import pytest

from aspire.flows import get_flow_wrapper


def test_get_flow_wrapper_zuko():
    FlowClass, xp = get_flow_wrapper(backend="zuko", flow_matching=False)
    import array_api_compat.torch as torch_api

    from aspire.flows.torch.flows import ZukoFlow

    assert FlowClass is ZukoFlow
    assert xp is torch_api


def test_get_flow_wrapper_zuko_flow_matching():
    FlowClass, xp = get_flow_wrapper(backend="zuko", flow_matching=True)
    import array_api_compat.torch as torch_api

    from aspire.flows.torch.flows import ZukoFlowMatching

    assert FlowClass is ZukoFlowMatching
    assert xp is torch_api


def test_get_flow_wrapper_flowjax():
    FlowClass, xp = get_flow_wrapper(backend="flowjax", flow_matching=False)
    import jax.numpy as jnp

    from aspire.flows.jax.flows import FlowJax

    assert FlowClass is FlowJax
    assert xp is jnp


def test_get_flow_wrapper_flowjax_flow_matching_not_implemented():
    with pytest.raises(
        NotImplementedError,
        match="Flow matching not implemented for JAX backend",
    ):
        get_flow_wrapper(backend="flowjax", flow_matching=True)


def test_get_flow_wrapper_external_backend_not_found():
    msg = "Unknown backend 'unknown_backend'. Available backends: "
    with pytest.raises(ValueError, match=msg):
        get_flow_wrapper(backend="unknown_backend")



import array_api_compat.numpy as np_xp
import array_api_compat.torch as torch_xp
import jax.numpy as jnp
import pytest

from aspire.utils import convert_dtype, resolve_dtype


def _dtype_name(dtype):
    if hasattr(dtype, "name") and dtype.name:
        return dtype.name
    if hasattr(dtype, "__name__"):
        return dtype.__name__
    text = str(dtype)
    if text.startswith("<class '") and text.endswith("'>"):
        text = text.split("'")[1]
    if text.startswith("dtype(") and text.endswith(")"):
        text = text[6:-1].strip("'\" ")
    return text.split(".")[-1]


@pytest.mark.parametrize(
    ("source_dtype", "target_xp", "expected"),
    [
        (np_xp.dtype("float32"), torch_xp, torch_xp.float32),
        (torch_xp.float64, np_xp, np_xp.dtype("float64")),
        (jnp.float16, torch_xp, torch_xp.float16),
        ("float32", jnp, jnp.dtype("float32")),
    ],
)
def test_convert_dtype_cross_namespace(source_dtype, target_xp, expected):
    converted = convert_dtype(source_dtype, target_xp)
    assert _dtype_name(converted) == _dtype_name(expected)


def test_convert_dtype_same_namespace_returns_original():
    dtype = np_xp.dtype("float32")
    converted = convert_dtype(dtype, np_xp)
    assert converted == dtype


def test_convert_dtype_none_returns_none():
    assert convert_dtype(None, torch_xp) is None


def test_convert_dtype_invalid_raises():
    with pytest.raises(ValueError):
        convert_dtype("not_a_real_dtype", torch_xp)


@pytest.mark.parametrize(
    ("value", "xp", "expected_name"),
    [
        ("float32", np_xp, "float32"),
        ("float64", torch_xp, "float64"),
        ("float16", jnp, "float16"),
        (np_xp.dtype("int32"), np_xp, "int32"),
        (torch_xp.float32, torch_xp, "float32"),
        (jnp.dtype("float64"), jnp, "float64"),
    ],
)
def test_resolve_dtype_matches_namespace(value, xp, expected_name):
    resolved = resolve_dtype(value, xp)
    assert _dtype_name(resolved) == expected_name


def test_resolve_dtype_passes_through_torch_dtype():
    dtype = torch_xp.float32
    assert resolve_dtype(dtype, torch_xp) is dtype


def test_resolve_dtype_errors_on_unknown():
    with pytest.raises(ValueError):
        resolve_dtype("not_a_real_dtype", np_xp)



# test_samples.py
import math
import pickle

import numpy as np
import pytest

from aspire.samples import BaseSamples, Samples, SMCSamples


def make_simple_samples(n=5, d=2, a=1.234):
    """
    Construct a simple, consistent set of arrays:
    - x: shape (n, d)
    - log_likelihood = a (constant)
    - log_prior = 0
    - log_q = 0
    This makes the math easy: evidence = exp(a), ESS = n, etc.
    """
    x = np.arange(n * d, dtype=float).reshape(n, d)
    ll = np.full(n, a, dtype=float)
    lp = np.zeros(n, dtype=float)
    lq = np.zeros(n, dtype=float)
    return x, ll, lp, lq


# -------- BaseSamples ---------------------------------------------------------


def test_basesamples_defaults_and_dims_numpy_namespace():
    x = np.zeros((3, 4))
    bs = BaseSamples(x=x)
    # xp inferred from x
    assert bs.xp is not None
    assert bs.dtype == x.dtype
    assert bs.x.dtype == x.dtype
    # parameters default names
    assert bs.parameters == ["x_0", "x_1", "x_2", "x_3"]
    # dims computed correctly
    assert bs.dims == 4
    # device set for numpy
    assert bs.device == "cpu"
    # __len__
    assert len(bs) == 3


def test_basesamples_to_dict_flat_and_nested():
    x = np.arange(6.0).reshape(3, 2)
    ll = np.array([0.1, 0.2, 0.3])
    bs = BaseSamples(x=x, log_likelihood=ll, parameters=["a", "b"])

    assert bs.log_likelihood.dtype == bs.dtype

    d_flat = bs.to_dict(flat=True)
    assert "a" in d_flat and "b" in d_flat
    assert "log_likelihood" in d_flat and np.allclose(
        d_flat["log_likelihood"], ll
    )

    d_nested = bs.to_dict(flat=False)
    assert "samples" in d_nested and isinstance(d_nested["samples"], dict)
    assert set(d_nested["samples"].keys()) == {"a", "b"}


def test_basesamples_to_numpy_and_namespace_roundtrip():
    x = np.arange(6.0).reshape(3, 2)
    ll = np.array([0.1, 0.2, 0.3])
    bs = BaseSamples(x=x, log_likelihood=ll, parameters=["a", "b"])

    bs_np = bs.to_numpy()
    assert isinstance(bs_np.x, np.ndarray)
    assert isinstance(bs_np.log_likelihood, np.ndarray)

    # Convert to the same namespace (numpy) explicitly
    bs_ns = bs.to_namespace(np)
    assert isinstance(bs_ns.x, np.ndarray)
    assert isinstance(bs_ns.log_likelihood, np.ndarray)


def test_basesamples_getitem_slice_returns_same_type_and_fields():
    x = np.arange(12.0).reshape(6, 2)
    ll = np.linspace(0, 1, 6)
    lp = np.linspace(1, 2, 6)
    lq = np.linspace(-1, 0, 6)
    bs = BaseSamples(
        x=x, log_likelihood=ll, log_prior=lp, log_q=lq, parameters=["a", "b"]
    )

    one = bs[2]
    assert isinstance(one, BaseSamples.__mro__[0]) or isinstance(
        one, BaseSamples
    )  # basic sanity
    assert one.x.shape == (2,)  # row
    assert one.parameters == ["a", "b"]
    assert np.isclose(one.log_likelihood, ll[2])
    assert np.isclose(one.log_prior, lp[2])
    assert np.isclose(one.log_q, lq[2])
    assert one.dtype == bs.dtype


def test_basesamples_respects_explicit_dtype_string():
    x = np.arange(6).reshape(3, 2)
    ll = np.linspace(0.0, 1.0, 3)
    bs = BaseSamples(x=x, log_likelihood=ll, dtype="float32")
    assert bs.x.dtype == np.dtype("float32")
    assert bs.dtype == np.dtype("float32")
    assert bs.log_likelihood.dtype == np.dtype("float32")


def test_basesamples_pickle_restores_namespace_and_fields():
    x = np.arange(6.0).reshape(3, 2)
    bs = BaseSamples(x=x)
    blob = pickle.dumps(bs)
    bs2 = pickle.loads(blob)
    # xp restored from x
    assert bs2.xp is not None
    # functionality intact
    assert bs2.dims == 2
    assert np.allclose(bs2.x, x)


# -------- Samples -------------------------------------------------------------


def test_samples_compute_weights_constant_case():
    n, d = 10, 3
    a = 1.5
    x, ll, lp, lq = make_simple_samples(n=n, d=d, a=a)
    s = Samples(
        x=x,
        log_likelihood=ll,
        log_prior=lp,
        log_q=lq,
        parameters=[f"p{i}" for i in range(d)],
    )

    # log_w = a; evidence = exp(a); evidence_error = 0; ESS = n
    assert np.allclose(s.log_w, a)
    assert math.isclose(float(s.evidence), math.exp(a), rel_tol=1e-12)
    assert math.isclose(float(s.log_evidence), a, rel_tol=1e-12)
    assert math.isclose(float(s.evidence_error), 0.0, abs_tol=1e-15)
    assert math.isclose(float(s.log_evidence_error), 0.0, abs_tol=1e-15)
    assert math.isclose(float(s.effective_sample_size), n, rel_tol=1e-12)
    assert math.isclose(float(s.efficiency), 1.0, rel_tol=1e-12)

    # scaled_weights ∝ exp(log_w - max)
    sw = s.scaled_weights
    assert np.allclose(sw, np.exp(a - a))  # all ones


def test_samples_to_dict_includes_weight_fields():
    x, ll, lp, lq = make_simple_samples(n=4, d=2, a=0.3)
    s = Samples(
        x=x, log_likelihood=ll, log_prior=lp, log_q=lq, parameters=["a", "b"]
    )
    d = s.to_dict(flat=True)
    for k in [
        "log_w",
        "weights",
        "evidence",
        "log_evidence",
        "evidence_error",
        "log_evidence_error",
        "effective_sample_size",
        "a",
        "b",
    ]:
        assert k in d


def test_samples_getitem_preserves_evidence_fields():
    x, ll, lp, lq = make_simple_samples(n=6, d=2, a=0.2)
    s = Samples(
        x=x, log_likelihood=ll, log_prior=lp, log_q=lq, parameters=["a", "b"]
    )
    s2 = s[1:4]
    assert isinstance(s2, Samples)
    assert s2.x.shape[0] == 3
    # evidence fields were set on s; __getitem__ preserves them via from_samples
    assert math.isclose(
        float(s2.log_evidence), float(s.log_evidence), rel_tol=1e-12
    )
    assert math.isclose(
        float(s2.log_evidence_error),
        float(s.log_evidence_error),
        rel_tol=1e-12,
    )


def test_samples_concatenate_success_and_mismatch_errors():
    x1, ll1, lp1, lq1 = make_simple_samples(n=3, d=2, a=0.0)
    x2, ll2, lp2, lq2 = make_simple_samples(n=4, d=2, a=0.1)
    s1 = Samples(
        x=x1,
        log_likelihood=ll1,
        log_prior=lp1,
        log_q=lq1,
        parameters=["a", "b"],
    )
    s2 = Samples(
        x=x2,
        log_likelihood=ll2,
        log_prior=lp2,
        log_q=lq2,
        parameters=["a", "b"],
    )

    sc = Samples.concatenate([s1, s2])
    assert isinstance(sc, Samples)
    assert sc.x.shape == (7, 2)
    assert sc.parameters == ["a", "b"]

    # Mismatching dtype -> error
    x3 = x1.astype(np.float32)
    s3 = Samples(
        x=x3,
        log_likelihood=ll1,
        log_prior=lp1,
        log_q=lq1,
        parameters=["a", "b"],
        dtype="float32",
    )
    with pytest.raises(ValueError):
        Samples.concatenate([s1, s3])

    # Mismatching parameters -> error
    t_bad = Samples(
        x=x2,
        log_likelihood=ll2,
        log_prior=lp2,
        log_q=lq2,
        parameters=["a", "c"],
    )
    with pytest.raises(ValueError):
        Samples.concatenate([s1, t_bad])

    # Mismatching namespaces -> error (simulate by converting to a different namespace if available)
    # Here both are numpy, so simulate by monkeypatching xp check with a subclass of BaseSamples? Easiest:
    # Create a dummy with same params but fake xp attr.
    s_fake = Samples(
        x=x2,
        log_likelihood=ll2,
        log_prior=lp2,
        log_q=lq2,
        parameters=["a", "b"],
    )
    s_fake.xp = object()  # break equality
    with pytest.raises(ValueError):
        Samples.concatenate([s1, s_fake])


def test_samples_rejection_sample_accepts_proportionally():
    # Make one point much heavier
    n, d = 100, 1
    x = np.arange(n, dtype=float).reshape(n, d)
    ll = np.zeros(n)
    lp = np.zeros(n)
    lq = np.zeros(n)
    ll[0] = 10.0  # very large log-likelihood for index 0
    s = Samples(
        x=x, log_likelihood=ll, log_prior=lp, log_q=lq, parameters=["p0"]
    )
    rng = np.random.default_rng(1234)
    rs = s.rejection_sample(rng=rng)
    assert isinstance(rs, Samples)
    assert len(rs) <= len(s)
    # Heaviest point should be present with high probability
    if len(rs) > 0:
        assert 0 in rs.x.squeeze().astype(int)


def test_smc_unnormalized_and_normalized_log_weights_and_ratio():
    n, d = 6, 2
    x = np.arange(n * d, dtype=float).reshape(n, d)
    ll = np.linspace(0.0, 1.0, n)
    lp = np.zeros(n)
    lq = np.zeros(n)
    beta0 = 0.3
    smc = SMCSamples(
        x=x,
        log_likelihood=ll,
        log_prior=lp,
        log_q=lq,
        beta=beta0,
        parameters=["a", "b"],
    )

    beta1 = 0.7
    # unnormalized weights (beta-beta0) * (ll+lp) since lq=0
    uw = smc.unnormalized_log_weights(beta1)
    assert np.allclose(uw, (beta1 - beta0) * (ll + lp))

    # log_evidence_ratio = logsumexp(uw) - log(n)
    import scipy.special as sps  # optional; if unavailable, reimplement quickly

    lse = sps.logsumexp(uw) - math.log(n)
    assert math.isclose(
        float(smc.log_evidence_ratio(beta1)), float(lse), rel_tol=1e-12
    )

    # normalized (shifted) log weights should not contain NaN and sum of exp(log_w - logsumexp(log_w)) = 1
    lw = smc.log_weights(beta1)
    w = np.exp(lw - sps.logsumexp(lw))
    assert np.isfinite(w).all()
    assert math.isclose(w.sum(), 1.0, rel_tol=1e-12)


def test_smc_log_weights_nan_guard_raises():
    x = np.zeros((3, 1))
    ll = np.array([0.0, np.nan, 1.0])
    smc = SMCSamples(
        x=x,
        log_likelihood=ll,
        log_prior=np.zeros(3),
        log_q=np.zeros(3),
        beta=0.0,
    )
    with pytest.raises(ValueError):
        _ = smc.log_weights(0.5)


def test_smc_resample_shapes_and_beta_and_warning_on_same_beta(caplog):
    n, d = 10, 2
    x = np.arange(n * d, dtype=float).reshape(n, d)
    ll = np.linspace(0, 1, n)
    lp = np.zeros(n)
    lq = np.zeros(n)
    beta0 = 0.2
    smc = SMCSamples(
        x=x,
        log_likelihood=ll,
        log_prior=lp,
        log_q=lq,
        beta=beta0,
        parameters=["a", "b"],
    )

    rng = np.random.default_rng(42)
    out = smc.resample(beta=0.8, n_samples=7, rng=rng)
    assert isinstance(out, SMCSamples)
    assert out.x.shape == (7, d)
    assert math.isclose(out.beta, 0.8)

    # same beta & n_samples=None -> returns self and logs a warning
    caplog.clear()
    with caplog.at_level("WARNING"):
        same = smc.resample(beta=beta0, n_samples=None, rng=rng)
    assert same is smc
    assert any(
        "Resampling with the same beta value" in r.message
        for r in caplog.records
    )


def test_smc_variance_delta_method_nonnegative():
    x = np.arange(10.0).reshape(5, 2)
    ll = np.linspace(0, 1, 5)
    smc = SMCSamples(
        x=x,
        log_likelihood=ll,
        log_prior=np.zeros(5),
        log_q=np.zeros(5),
        beta=0.3,
    )
    var = smc.log_evidence_ratio_variance(beta=0.6)
    assert float(var) >= 0.0 or np.isnan(var)


def test_smc_to_standard_samples_and_getitem_preserve_fields():
    x = np.arange(8.0).reshape(4, 2)
    ll = np.linspace(0, 1, 4)
    smc = SMCSamples(
        x=x,
        log_likelihood=ll,
        log_prior=np.zeros(4),
        log_q=np.zeros(4),
        beta=0.4,
        parameters=["a", "b"],
        log_evidence=1.23,
    )

    std = smc.to_standard_samples()
    assert isinstance(std, Samples)
    # For std Samples, weights aren't computed (log_q missing by design),
    # but log_evidence is carried over.
    assert std.log_w is None
    assert math.isclose(float(std.log_evidence), 1.23, rel_tol=1e-12)

    # __getitem__ preserves beta and log_evidence
    smc2 = smc[1:3]
    assert isinstance(smc2, SMCSamples)
    assert smc2.x.shape[0] == 2
    assert math.isclose(float(smc2.beta), 0.4, rel_tol=1e-12)
    assert math.isclose(float(smc2.log_evidence), 1.23, rel_tol=1e-12)


def test_str_contains_counts_and_metrics():
    x, ll, lp, lq = make_simple_samples(n=4, d=2, a=0.7)
    s = Samples(
        x=x, log_likelihood=ll, log_prior=lp, log_q=lq, parameters=["a", "b"]
    )
    s_str = str(s)
    assert "No. samples: 4" in s_str
    assert "No. parameters: 2" in s_str
    assert "Log evidence:" in s_str
    assert "Effective sample size:" in s_str
    assert "Efficiency:" in s_str



"""Example using sequential posterior inference with SMC.

This example uses JAX for computations and BlackJAX for the MCMC sampling
in SMC step.
"""

from pathlib import Path

import jax
import jax.numpy as jnp

from aspire import Aspire
from aspire.plot import plot_comparison
from aspire.samples import Samples
from aspire.utils import configure_logger

# RNG for generating initial samples
key = jax.random.key(42)

# Output directory
outdir = Path("outdir") / "blackjax_smc_example"
outdir.mkdir(parents=True, exist_ok=True)

# Configure logger to show INFO level messages
configure_logger()

# Number of dimensions
dims = 4
# Parameter names
parameters = [f"x{i}" for i in range(dims)]
# Prior bounds
prior_bounds = {param: (-5, 5) for param in parameters}

# Means and covariances of the two Gaussian components
mu1 = 2 * jnp.ones(dims)
mu2 = -2 * jnp.ones(dims)
cov1 = 0.5 * jnp.eye(dims)
cov2 = jnp.eye(dims)


def log_likelihood(samples):
    """Log-likelihood of a mixture of two Gaussians"""
    x = samples.x
    comp1 = (
        -0.5 * ((x - mu1) @ jnp.linalg.inv(cov1) * (x - mu1)).sum(axis=-1)
        - 0.5 * dims * jnp.log(2 * jnp.pi)
        - 0.5 * jnp.linalg.slogdet(cov1)[1]
    )
    comp2 = (
        -0.5 * ((x - mu2) @ jnp.linalg.inv(cov2) * (x - mu2)).sum(axis=-1)
        - 0.5 * dims * jnp.log(2 * jnp.pi)
        - 0.5 * jnp.linalg.slogdet(cov2)[1]
    )
    return jnp.logaddexp(comp1, comp2)  # Log-sum-exp for numerical stability


def log_prior(samples):
    """Uniform prior between -5 and 5 in each dimension"""
    x = samples.x
    in_bounds = jnp.all((x >= -5) & (x <= 5), axis=-1)
    logp = jnp.where(in_bounds, -dims * jnp.log(10), -jnp.inf)
    return logp


# Generate prior samples for comparison, these are not used in SMC
key, prior_key = jax.random.split(key)
prior_samples = Samples(
    jax.random.uniform(prior_key, shape=(5000, dims), minval=-5, maxval=5),
    parameters=parameters,
)

# True posterior samples for comparison
key, post_key0, post_key1 = jax.random.split(key, 3)
true_posterior_samples = Samples(
    jnp.concatenate(
        [
            jax.random.multivariate_normal(
                post_key0, mu1, cov1, shape=(2500,)
            ),
            jax.random.multivariate_normal(
                post_key1, mu2, cov2, shape=(2500,)
            ),
        ],
        axis=0,
    ),
    parameters=parameters,
)

# We draw initial samples from two Gaussians centered away from the true modes
# to demonstrate the ability of SMC to explore the posterior
key, offset_key1, offset_key2 = jax.random.split(key, 3)
offset_1 = jax.random.uniform(offset_key1, shape=(dims,), minval=-3, maxval=3)
offset_2 = jax.random.uniform(offset_key2, shape=(dims,), minval=-3, maxval=3)
key, init_key1, init_key2 = jax.random.split(key, 3)
initial_samples = jnp.concatenate(
    [
        jax.random.normal(init_key1, shape=(2500, dims)) + (mu1 - offset_1),
        jax.random.normal(init_key2, shape=(2500, dims)) + (mu2 - offset_2),
    ],
    axis=0,
)
initial_samples = Samples(initial_samples, parameters=parameters)

# Initialize Aspire with the log-likelihood and log-prior
key, aspire_key = jax.random.split(key)
aspire = Aspire(
    log_likelihood=log_likelihood,
    log_prior=log_prior,
    dims=dims,
    flow_backend="flowjax",  # Use Flowjax as the normalizing flow backend
    prior_bounds=prior_bounds,  # Specify prior bounds
    parameters=parameters,
    key=aspire_key,
)

# Fit the normalizing flow to the initial samples
fit_history = aspire.fit(initial_samples, max_epochs=30)

# Plot loss
fit_history.plot_loss().savefig(outdir / "loss.png")

# Sample from the posterior using SMC
# We use BlackJAX's NUTS as the MCMC kernel within SMC
# We enable the bounded to unbounded transform in the preconditioning to avoid
# issues with NUTS on bounded spaces
samples, history = aspire.sample_posterior(
    sampler="blackjax_smc",  # use the BlackJAX SMC sampler
    n_samples=500,  # Number of particles in SMC
    n_final_samples=5000,  # Number of samples to draw from the final distribution
    adaptive=True,
    target_efficiency=0.8,
    sampler_kwargs=dict(  # Keyword arguments for the specific sampler
        algorithm="nuts",  # Use NUTS within SMC
        step_size=0.1,  # Step size for NUTS, this will need tuning
        n_steps=20,  # Number of leapfrog steps for NUTS
    ),
    preconditioning_kwargs=dict(
        affine_transform=True,  # Use affine transform preconditioning
        bounded_to_unbounded=True,  # Transform bounded parameters to unbounded space
    ),
    return_history=True,  # To return the SMC history (e.g., ESS, betas)
)
# Plot SMC diagnostics
history.plot().savefig(outdir / "smc_diagnostics.png")

# Plot corner plot of the samples
# Include initial samples and prior samples for comparison
plot_comparison(
    initial_samples,
    true_posterior_samples,
    samples,
    labels=["Initial Samples", "True Posterior Samples", "SMC Samples"],
    per_samples_kwargs=[
        {"color": "grey"},
        {"color": "k"},
        {"color": "C1"},
    ],
).savefig(outdir / "posterior.png")



import logging
from typing import Any

logger = logging.getLogger(__name__)


def get_flow_wrapper(
    backend: str = "zuko", flow_matching: bool = False
) -> tuple[type, Any]:
    """Get the wrapper for the flow implementation.

    Parameters
    ----------
    backend : str
        The backend to use. Options are "zuko" (PyTorch), "flowjax" (JAX), or
        any other registered flow class via entry points. Default is "zuko".
    flow_matching : bool, optional
        Whether to use flow matching variant of the flow. Default is False.

    Returns
    -------
    FlowClass : type
        The flow class corresponding to the specified backend.
    xp : module
        The array API module corresponding to the specified backend.
    """
    from importlib.metadata import entry_points

    if backend == "zuko":
        import array_api_compat.torch as torch_api

        from .torch.flows import ZukoFlow, ZukoFlowMatching

        if flow_matching:
            return ZukoFlowMatching, torch_api
        else:
            return ZukoFlow, torch_api
    elif backend == "flowjax":
        import jax.numpy as jnp

        from .jax.flows import FlowJax

        if flow_matching:
            raise NotImplementedError(
                "Flow matching not implemented for JAX backend"
            )
        return FlowJax, jnp
    else:
        if flow_matching:
            logger.warning(
                "Flow matching option is ignored for external backends."
            )
        eps = {
            ep.name.lower(): ep for ep in entry_points(group="aspire.flows")
        }
        if backend in eps:
            FlowClass = eps[backend].load()
            xp = getattr(FlowClass, "xp", None)
            if xp is None:
                raise ValueError(
                    f"Flow class {backend} does not define an `xp` attribute"
                )
            return FlowClass, xp
        else:
            known_backends = ["zuko", "flowjax"] + list(eps.keys())
            raise ValueError(
                f"Unknown backend '{backend}'. Available backends: {known_backends}"
            )



import logging
import multiprocessing as mp
from inspect import signature
from typing import Any, Callable

import h5py

from .flows import get_flow_wrapper
from .flows.base import Flow
from .history import History
from .samples import Samples
from .transforms import (
    CompositeTransform,
    FlowPreconditioningTransform,
    FlowTransform,
)
from .utils import recursively_save_to_h5_file

logger = logging.getLogger(__name__)


class Aspire:
    """Accelerated Sequential Posterior Inference via REuse (aspire).

    Parameters
    ----------
    log_likelihood : Callable
        The log likelihood function.
    log_prior : Callable
        The log prior function.
    dims : int
        The number of dimensions.
    parameters : list[str] | None
        The list of parameter names. If None, any samples objects will not
        have the parameters names specified.
    periodic_parameters : list[str] | None
        The list of periodic parameters.
    prior_bounds : dict[str, tuple[float, float]] | None
        The bounds for the prior. If None, some parameter transforms cannot
        be applied.
    bounded_to_unbounded : bool
        Whether to transform bounded parameters to unbounded ones.
    bounded_transform : str
        The transformation to use for bounded parameters. Options are
        'logit', 'exp', or 'tanh'.
    device : str | None
        The device to use for the flow. If None, the default device will be
        used. This is only used when using the PyTorch backend.
    xp : Callable | None
        The array backend to use. If None, the default backend will be
        used.
    flow : Flow | None
        The flow object, if it already exists.
        If None, a new flow will be created.
    flow_backend : str
        The backend to use for the flow. Options are 'zuko' or 'flowjax'.
    flow_matching : bool
        Whether to use flow matching.
    eps : float
        The epsilon value to use for data transforms.
    dtype : Any | str | None
        The data type to use for the samples, flow and transforms.
    **kwargs
        Keyword arguments to pass to the flow.
    """

    def __init__(
        self,
        *,
        log_likelihood: Callable,
        log_prior: Callable,
        dims: int,
        parameters: list[str] | None = None,
        periodic_parameters: list[str] | None = None,
        prior_bounds: dict[str, tuple[float, float]] | None = None,
        bounded_to_unbounded: bool = True,
        bounded_transform: str = "logit",
        device: str | None = None,
        xp: Callable | None = None,
        flow: Flow | None = None,
        flow_backend: str = "zuko",
        flow_matching: bool = False,
        eps: float = 1e-6,
        dtype: Any | str | None = None,
        **kwargs,
    ) -> None:
        self.log_likelihood = log_likelihood
        self.log_prior = log_prior
        self.dims = dims
        self.parameters = parameters
        self.device = device
        self.eps = eps

        self.periodic_parameters = periodic_parameters
        self.prior_bounds = prior_bounds
        self.bounded_to_unbounded = bounded_to_unbounded
        self.bounded_transform = bounded_transform
        self.flow_matching = flow_matching
        self.flow_backend = flow_backend
        self.flow_kwargs = kwargs
        self.xp = xp
        self.dtype = dtype

        self._flow = flow

    @property
    def flow(self):
        """The normalizing flow object."""
        return self._flow

    @flow.setter
    def flow(self, flow: Flow):
        """Set the normalizing flow object."""
        self._flow = flow

    @property
    def sampler(self):
        """The sampler object."""
        return self._sampler

    @property
    def n_likelihood_evaluations(self):
        """The number of likelihood evaluations."""
        if hasattr(self, "_sampler"):
            return self._sampler.n_likelihood_evaluations
        else:
            return None

    def convert_to_samples(
        self,
        x,
        log_likelihood=None,
        log_prior=None,
        log_q=None,
        evaluate: bool = True,
        xp=None,
    ) -> Samples:
        if xp is None:
            xp = self.xp
        samples = Samples(
            x=x,
            parameters=self.parameters,
            log_likelihood=log_likelihood,
            log_prior=log_prior,
            log_q=log_q,
            xp=xp,
            dtype=self.dtype,
        )

        if evaluate:
            if log_prior is None:
                logger.info("Evaluating log prior")
                samples.log_prior = samples.xp.to_device(
                    self.log_prior(samples), samples.device
                )
            if log_likelihood is None:
                logger.info("Evaluating log likelihood")
                samples.log_likelihood = samples.xp.to_device(
                    self.log_likelihood(samples), samples.device
                )
            samples.compute_weights()
        return samples

    def init_flow(self):
        FlowClass, xp = get_flow_wrapper(
            backend=self.flow_backend, flow_matching=self.flow_matching
        )

        data_transform = FlowTransform(
            parameters=self.parameters,
            prior_bounds=self.prior_bounds,
            bounded_to_unbounded=self.bounded_to_unbounded,
            bounded_transform=self.bounded_transform,
            device=self.device,
            xp=xp,
            eps=self.eps,
            dtype=self.dtype,
        )

        # Check if FlowClass takes `parameters` as an argument
        flow_init_params = signature(FlowClass.__init__).parameters
        if "parameters" in flow_init_params:
            self.flow_kwargs["parameters"] = self.parameters.copy()

        logger.info(f"Configuring {FlowClass} with kwargs: {self.flow_kwargs}")

        self._flow = FlowClass(
            dims=self.dims,
            device=self.device,
            data_transform=data_transform,
            dtype=self.dtype,
            **self.flow_kwargs,
        )

    def fit(self, samples: Samples, **kwargs) -> History:
        if self.xp is None:
            self.xp = samples.xp

        if self.flow is None:
            self.init_flow()

        self.training_samples = samples
        logger.info(f"Training with {len(samples.x)} samples")
        history = self.flow.fit(samples.x, **kwargs)
        return history

    def get_sampler_class(self, sampler_type: str) -> Callable:
        """Get the sampler class based on the sampler type.

        Parameters
        ----------
        sampler_type : str
            The type of sampler to use. Options are 'importance', 'emcee', or 'smc'.
        """
        if sampler_type == "importance":
            from .samplers.importance import ImportanceSampler as SamplerClass
        elif sampler_type == "emcee":
            from .samplers.mcmc import Emcee as SamplerClass
        elif sampler_type == "emcee_smc":
            from .samplers.smc.emcee import EmceeSMC as SamplerClass
        elif sampler_type == "minipcn":
            from .samplers.mcmc import MiniPCN as SamplerClass
        elif sampler_type in ["smc", "minipcn_smc"]:
            from .samplers.smc.minipcn import MiniPCNSMC as SamplerClass
        elif sampler_type == "blackjax_smc":
            from .samplers.smc.blackjax import BlackJAXSMC as SamplerClass
        else:
            raise ValueError(f"Unknown sampler type: {sampler_type}")
        return SamplerClass

    def init_sampler(
        self,
        sampler_type: str,
        preconditioning: str | None = None,
        preconditioning_kwargs: dict | None = None,
        **kwargs,
    ) -> Callable:
        """Initialize the sampler for posterior sampling.

        Parameters
        ----------
        sampler_type : str
            The type of sampler to use. Options are 'importance', 'emcee', or 'smc'.
        """
        SamplerClass = self.get_sampler_class(sampler_type)

        if sampler_type != "importance" and preconditioning is None:
            preconditioning = "default"

        preconditioning = preconditioning.lower() if preconditioning else None

        if preconditioning is None or preconditioning == "none":
            transform = None
        elif preconditioning in ["standard", "default"]:
            preconditioning_kwargs = preconditioning_kwargs or {}
            preconditioning_kwargs.setdefault("affine_transform", False)
            preconditioning_kwargs.setdefault("bounded_to_unbounded", False)
            preconditioning_kwargs.setdefault("bounded_transform", "logit")
            transform = CompositeTransform(
                parameters=self.parameters,
                prior_bounds=self.prior_bounds,
                periodic_parameters=self.periodic_parameters,
                xp=self.xp,
                device=self.device,
                dtype=self.dtype,
                **preconditioning_kwargs,
            )
        elif preconditioning == "flow":
            preconditioning_kwargs = preconditioning_kwargs or {}
            preconditioning_kwargs.setdefault("affine_transform", False)
            transform = FlowPreconditioningTransform(
                parameters=self.parameters,
                flow_backend=self.flow_backend,
                flow_kwargs=self.flow_kwargs,
                flow_matching=self.flow_matching,
                periodic_parameters=self.periodic_parameters,
                bounded_to_unbounded=self.bounded_to_unbounded,
                prior_bounds=self.prior_bounds,
                xp=self.xp,
                dtype=self.dtype,
                device=self.device,
                **preconditioning_kwargs,
            )
        else:
            raise ValueError(f"Unknown preconditioning: {preconditioning}")

        sampler = SamplerClass(
            log_likelihood=self.log_likelihood,
            log_prior=self.log_prior,
            dims=self.dims,
            prior_flow=self.flow,
            xp=self.xp,
            dtype=self.dtype,
            preconditioning_transform=transform,
            **kwargs,
        )
        return sampler

    def sample_posterior(
        self,
        n_samples: int = 1000,
        sampler: str = "importance",
        xp: Any = None,
        return_history: bool = False,
        preconditioning: str | None = None,
        preconditioning_kwargs: dict | None = None,
        **kwargs,
    ) -> Samples:
        """Draw samples from the posterior distribution.

        If using a sampler that calls an external sampler, e.g.
        :code:`minipcn` then keyword arguments for this sampler should be
        specified in :code:`sampler_kwargs`. For example:

        .. code-block:: python

            aspire = aspire(...)
            aspire.sample_posterior(
                n_samples=1000,
                sampler="minipcn_smc",
                adaptive=True,
                sampler_kwargs=dict(
                    n_steps=100,
                    step_fn="tpcn",
                )
            )

        Parameters
        ----------
        n_samples : int
            The number of sample to draw.
        sampler: str
            Sampling algorithm to use for drawing the posterior samples.
        xp: Any
            Array API for the final samples.
        return_history : bool
            Whether to return the history of the sampler.
        preconditioning: str
            Type of preconditioning to apply in the sampler. Options are
            'default', 'flow', or 'none'. If not specified, the default
            will depend on the sampler being used. The importance sampler
            will default to 'none' and the other samplers to 'default'
        preconditioning_kwargs: dict
            Keyword arguments to pass to the preconditioning transform.
        kwargs : dict
            Keyword arguments to pass to the sampler. These are passed
            automatically to the init method of the sampler or to the sample
            method.

        Returns
        -------
        samples : Samples
            Samples object contain samples and their corresponding weights.
        """
        SamplerClass = self.get_sampler_class(sampler)
        # Determine sampler initialization parameters
        # and remove them from kwargs
        sampler_init_kwargs = signature(SamplerClass.__init__).parameters
        sampler_kwargs = {
            k: v
            for k, v in kwargs.items()
            if k in sampler_init_kwargs and k != "self"
        }
        kwargs = {
            k: v
            for k, v in kwargs.items()
            if k not in sampler_init_kwargs or k == "self"
        }

        self._sampler = self.init_sampler(
            sampler,
            preconditioning=preconditioning,
            preconditioning_kwargs=preconditioning_kwargs,
            **sampler_kwargs,
        )
        samples = self._sampler.sample(n_samples, **kwargs)
        if xp is not None:
            samples = samples.to_namespace(xp)
        samples.parameters = self.parameters
        logger.info(f"Sampled {len(samples)} samples from the posterior")
        logger.info(
            f"Number of likelihood evaluations: {self.n_likelihood_evaluations}"
        )
        logger.info("Sample summary:")
        logger.info(samples)
        if return_history:
            return samples, self._sampler.history
        else:
            return samples

    def enable_pool(self, pool: mp.Pool, **kwargs):
        """Context manager to temporarily replace the log_likelihood method
        with a version that uses a multiprocessing pool to parallelize
        computation.

        Parameters
        ----------
        pool : multiprocessing.Pool
            The pool to use for parallel computation.
        """
        from .utils import PoolHandler

        return PoolHandler(self, pool, **kwargs)

    def config_dict(
        self, include_sampler_config: bool = True, **kwargs
    ) -> dict:
        """Return a dictionary with the configuration of the aspire object.

        Parameters
        ----------
        include_sampler_config : bool
            Whether to include the configuration of the sampler. Default is
            True.
        kwargs : dict
            Additional keyword arguments to pass to the :py:meth:`config_dict`
            method of the sampler.
        """
        config = {
            "log_likelihood": self.log_likelihood.__name__,
            "log_prior": self.log_prior.__name__,
            "dims": self.dims,
            "parameters": self.parameters,
            "periodic_parameters": self.periodic_parameters,
            "prior_bounds": self.prior_bounds,
            "bounded_to_unbounded": self.bounded_to_unbounded,
            "bounded_transform": self.bounded_transform,
            "flow_matching": self.flow_matching,
            "device": self.device,
            "xp": self.xp.__name__ if self.xp else None,
            "flow_backend": self.flow_backend,
            "flow_kwargs": self.flow_kwargs,
            "eps": self.eps,
        }
        if include_sampler_config:
            config["sampler_config"] = self.sampler.config_dict(**kwargs)
        return config

    def save_config(
        self, h5_file: h5py.File, path="aspire_config", **kwargs
    ) -> None:
        """Save the configuration to an HDF5 file.

        Parameters
        ----------
        h5_file : h5py.File
            The HDF5 file to save the configuration to.
        path : str
            The path in the HDF5 file to save the configuration to.
        kwargs : dict
            Additional keyword arguments to pass to the :py:meth:`config_dict`
            method.
        """
        recursively_save_to_h5_file(
            h5_file,
            path,
            self.config_dict(**kwargs),
        )

    def save_flow(self, h5_file: h5py.File, path="flow") -> None:
        """Save the flow to an HDF5 file.

        Parameters
        ----------
        h5_file : h5py.File
            The HDF5 file to save the flow to.
        path : str
            The path in the HDF5 file to save the flow to.
        """
        if self.flow is None:
            raise ValueError("Flow has not been initialized.")
        self.flow.save(h5_file, path=path)

    def load_flow(self, h5_file: h5py.File, path="flow") -> None:
        """Load the flow from an HDF5 file.

        Parameters
        ----------
        h5_file : h5py.File
            The HDF5 file to load the flow from.
        path : str
            The path in the HDF5 file to load the flow from.
        """
        FlowClass, xp = get_flow_wrapper(
            backend=self.flow_backend, flow_matching=self.flow_matching
        )
        self._flow = FlowClass.load(h5_file, path=path)

    def save_config_to_json(self, filename: str) -> None:
        """Save the configuration to a JSON file."""
        import json

        with open(filename, "w") as f:
            json.dump(self.config_dict(), f, indent=4)

    def sample_flow(self, n_samples: int = 1, xp=None) -> Samples:
        """Sample from the flow directly.

        Includes the data transform, but does not compute
        log likelihood or log prior.
        """
        if self.flow is None:
            self.init_flow()
        x, log_q = self.flow.sample_and_log_prob(n_samples)
        samples = Samples(x=x, log_q=log_q, xp=xp, parameters=self.parameters)
        return samples



"""
aspire: Accelerated Sequential Posterior Inference via REuse
"""

import logging
from importlib.metadata import PackageNotFoundError, version

from .aspire import Aspire
from .samples import Samples

try:
    __version__ = version("aspire")
except PackageNotFoundError:
    __version__ = "unknown"

logging.getLogger(__name__).addHandler(logging.NullHandler())

__all__ = [
    "Aspire",
    "Samples",
]



import numpy as np

from ..samples import Samples, to_numpy
from ..utils import track_calls
from .base import Sampler


class MCMCSampler(Sampler):
    def draw_initial_samples(self, n_samples: int) -> Samples:
        """Draw initial samples from the prior flow."""
        # Flow may propose samples outside prior bounds, so we may need
        # to try multiple times to get enough valid samples.
        n_samples_drawn = 0
        samples = None
        while n_samples_drawn < n_samples:
            n_to_draw = n_samples - n_samples_drawn
            x, log_q = self.prior_flow.sample_and_log_prob(n_to_draw)
            new_samples = Samples(x, xp=self.xp, log_q=log_q, dtype=self.dtype)
            new_samples.log_prior = new_samples.array_to_namespace(
                self.log_prior(new_samples)
            )
            valid = self.xp.isfinite(new_samples.log_prior)
            n_valid = self.xp.sum(valid)
            if n_valid > 0:
                if samples is None:
                    samples = new_samples[valid]
                else:
                    samples = Samples.concatenate(
                        [samples, new_samples[valid]]
                    )
                n_samples_drawn += n_valid

        if n_samples_drawn > n_samples:
            samples = samples[:n_samples]

        samples.log_likelihood = samples.array_to_namespace(
            self.log_likelihood(samples)
        )
        return samples

    def log_prob(self, z):
        """Compute the log probability of the samples.

        Input samples are in the transformed space.
        """
        x, log_abs_det_jacobian = self.preconditioning_transform.inverse(z)
        samples = Samples(x, xp=self.xp, dtype=self.dtype)
        samples.log_prior = self.log_prior(samples)
        samples.log_likelihood = self.log_likelihood(samples)
        log_prob = (
            samples.log_likelihood
            + samples.log_prior
            + samples.array_to_namespace(log_abs_det_jacobian)
        )
        return to_numpy(log_prob).flatten()


class Emcee(MCMCSampler):
    @track_calls
    def sample(
        self,
        n_samples: int,
        nwalkers: int = None,
        nsteps: int = 500,
        rng=None,
        discard=0,
        **kwargs,
    ) -> Samples:
        from emcee import EnsembleSampler

        nwalkers = nwalkers or n_samples
        self.sampler = EnsembleSampler(
            nwalkers,
            self.dims,
            log_prob_fn=self.log_prob,
            vectorize=True,
        )

        rng = rng or np.random.default_rng()

        samples = self.draw_initial_samples(nwalkers)
        p0 = samples.x

        z0 = to_numpy(self.preconditioning_transform.fit(p0))

        self.sampler.run_mcmc(z0, nsteps, **kwargs)

        z = self.sampler.get_chain(flat=True, discard=discard)
        x = self.preconditioning_transform.inverse(z)[0]

        x_evidence, log_q = self.prior_flow.sample_and_log_prob(n_samples)
        samples_evidence = Samples(x_evidence, log_q=log_q, xp=self.xp)
        samples_evidence.log_prior = self.log_prior(samples_evidence)
        samples_evidence.log_likelihood = self.log_likelihood(samples_evidence)
        samples_evidence.compute_weights()

        samples_mcmc = Samples(
            x, xp=self.xp, parameters=self.parameters, dtype=self.dtype
        )
        samples_mcmc.log_prior = samples_mcmc.array_to_namespace(
            self.log_prior(samples_mcmc)
        )
        samples_mcmc.log_likelihood = samples_mcmc.array_to_namespace(
            self.log_likelihood(samples_mcmc)
        )
        samples_mcmc.log_evidence = samples_mcmc.array_to_namespace(
            samples_evidence.log_evidence
        )
        samples_mcmc.log_evidence_error = samples_mcmc.array_to_namespace(
            samples_evidence.log_evidence_error
        )

        return samples_mcmc


class MiniPCN(MCMCSampler):
    @track_calls
    def sample(
        self,
        n_samples,
        rng=None,
        target_acceptance_rate=0.234,
        n_steps=100,
        thin=1,
        burnin=0,
        last_step_only=False,
        step_fn="tpcn",
    ):
        from minipcn import Sampler

        rng = rng or np.random.default_rng()
        p0 = self.draw_initial_samples(n_samples).x

        z0 = to_numpy(self.preconditioning_transform.fit(p0))

        self.sampler = Sampler(
            log_prob_fn=self.log_prob,
            step_fn=step_fn,
            rng=rng,
            dims=self.dims,
            target_acceptance_rate=target_acceptance_rate,
        )

        chain, history = self.sampler.sample(z0, n_steps=n_steps)

        if last_step_only:
            z = chain[-1]
        else:
            z = chain[burnin::thin].reshape(-1, self.dims)

        x = self.preconditioning_transform.inverse(z)[0]

        samples_mcmc = Samples(x, xp=self.xp, parameters=self.parameters)
        samples_mcmc.log_prior = samples_mcmc.array_to_namespace(
            self.log_prior(samples_mcmc)
        )
        samples_mcmc.log_likelihood = samples_mcmc.array_to_namespace(
            self.log_likelihood(samples_mcmc)
        )
        return samples_mcmc



from __future__ import annotations

import copy
from dataclasses import dataclass, field

import matplotlib.pyplot as plt
from matplotlib.figure import Figure

from .utils import recursively_save_to_h5_file


@dataclass
class History:
    """Base class for storing history of a sampler."""

    def save(self, h5_file, path="history"):
        """Save the history to an HDF5 file."""
        dictionary = copy.deepcopy(self.__dict__)
        recursively_save_to_h5_file(h5_file, path, dictionary)


@dataclass
class FlowHistory(History):
    training_loss: list[float] = field(default_factory=list)
    validation_loss: list[float] = field(default_factory=list)

    def plot_loss(self) -> Figure:
        """Plot the training and validation loss."""
        fig = plt.figure()
        plt.plot(self.training_loss, label="Training loss")
        plt.plot(self.validation_loss, label="Validation loss")
        plt.legend()
        plt.xlabel("Epoch")
        plt.ylabel("Loss")
        return fig

    def save(self, h5_file, path="flow_history"):
        """Save the history to an HDF5 file."""
        super().save(h5_file, path=path)


@dataclass
class SMCHistory(History):
    log_norm_ratio: list[float] = field(default_factory=list)
    log_norm_ratio_var: list[float] = field(default_factory=list)
    beta: list[float] = field(default_factory=list)
    ess: list[float] = field(default_factory=list)
    ess_target: list[float] = field(default_factory=list)
    eff_target: list[float] = field(default_factory=list)
    mcmc_autocorr: list[float] = field(default_factory=list)
    mcmc_acceptance: list[float] = field(default_factory=list)

    def save(self, h5_file, path="smc_history"):
        """Save the history to an HDF5 file."""
        super().save(h5_file, path=path)

    def plot_beta(self, ax=None) -> Figure | None:
        if ax is None:
            fig, ax = plt.subplots()
        else:
            fig = None
        ax.plot(self.beta)
        ax.set_xlabel("Iteration")
        ax.set_ylabel(r"$\beta$")
        return fig

    def plot_log_norm_ratio(self, ax=None) -> Figure | None:
        if ax is None:
            fig, ax = plt.subplots()
        else:
            fig = None
        ax.plot(self.log_norm_ratio)
        ax.set_xlabel("Iteration")
        ax.set_ylabel("Log evidence ratio")
        return fig

    def plot_ess(self, ax=None) -> Figure | None:
        if ax is None:
            fig, ax = plt.subplots()
        else:
            fig = None
        ax.plot(self.ess)
        ax.set_xlabel("Iteration")
        ax.set_ylabel("ESS")
        return fig

    def plot_ess_target(self, ax=None) -> Figure | None:
        if ax is None:
            fig, ax = plt.subplots()
        else:
            fig = None
        ax.plot(self.ess_target)
        ax.set_xlabel("Iteration")
        ax.set_ylabel("ESS target")
        return fig

    def plot_eff_target(self, ax=None) -> Figure | None:
        if ax is None:
            fig, ax = plt.subplots()
        else:
            fig = None
        ax.plot(self.eff_target)
        ax.set_xlabel("Iteration")
        ax.set_ylabel("Efficiency target")
        return fig

    def plot_mcmc_acceptance(self, ax=None) -> Figure | None:
        if ax is None:
            fig, ax = plt.subplots()
        else:
            fig = None
        ax.plot(self.mcmc_acceptance)
        ax.set_xlabel("Iteration")
        ax.set_ylabel("MCMC Acceptance")
        return fig

    def plot_mcmc_autocorr(self, ax=None) -> Figure | None:
        if ax is None:
            fig, ax = plt.subplots()
        else:
            fig = None
        ax.plot(self.mcmc_autocorr)
        ax.set_xlabel("Iteration")
        ax.set_ylabel("MCMC Autocorr")
        return fig

    def plot(self, fig: Figure | None = None) -> Figure:
        methods = [
            self.plot_beta,
            self.plot_log_norm_ratio,
            self.plot_ess,
            self.plot_ess_target,
            self.plot_eff_target,
            self.plot_mcmc_acceptance,
        ]

        if fig is None:
            fig, axs = plt.subplots(len(methods), 1, sharex=True)
        else:
            axs = fig.axes

        for method, ax in zip(methods, axs):
            method(ax)

        for ax in axs[:-1]:
            ax.set_xlabel("")

        return fig



from __future__ import annotations

import copy
import logging
import math
from dataclasses import dataclass, field
from typing import Any, Callable

import numpy as np
from array_api_compat import (
    array_namespace,
    is_numpy_namespace,
    to_device,
)
from array_api_compat import device as api_device
from array_api_compat.common._typing import Array
from matplotlib.figure import Figure

from .utils import (
    asarray,
    convert_dtype,
    logsumexp,
    recursively_save_to_h5_file,
    resolve_dtype,
    to_numpy,
)

logger = logging.getLogger(__name__)


@dataclass
class BaseSamples:
    """Class for storing samples and corresponding weights.

    If :code:`xp` is not specified, all inputs will be converted to match
    the array type of :code:`x`.
    """

    x: Array
    """Array of samples, shape (n_samples, n_dims)."""
    log_likelihood: Array | None = None
    """Log-likelihood values for the samples."""
    log_prior: Array | None = None
    """Log-prior values for the samples."""
    log_q: Array | None = None
    """Log-probability values under the proposal distribution."""
    parameters: list[str] | None = None
    """List of parameter names."""
    dtype: Any | str | None = None
    """Data type of the samples.

    If None, the default dtype for the array namespace will be used.
    """
    xp: Callable | None = None
    """
    The array namespace to use for the samples.

    If None, the array namespace will be inferred from the type of :code:`x`.
    """
    device: Any = None
    """Device to store the samples on.

    If None, the device will be inferred from the array namespace of :code:`x`.
    """

    def __post_init__(self):
        if self.xp is None:
            self.xp = array_namespace(self.x)
        # Numpy arrays need to be on the CPU before being converted
        if is_numpy_namespace(self.xp):
            self.device = "cpu"
        if self.dtype is not None:
            self.dtype = resolve_dtype(self.dtype, self.xp)
        else:
            # Fall back to default dtype for the array namespace
            self.dtype = None
        self.x = self.array_to_namespace(self.x, dtype=self.dtype)
        if self.device is None:
            self.device = api_device(self.x)
        if self.log_likelihood is not None:
            self.log_likelihood = self.array_to_namespace(
                self.log_likelihood, dtype=self.dtype
            )
        if self.log_prior is not None:
            self.log_prior = self.array_to_namespace(
                self.log_prior, dtype=self.dtype
            )
        if self.log_q is not None:
            self.log_q = self.array_to_namespace(self.log_q, dtype=self.dtype)

        if self.parameters is None:
            self.parameters = [f"x_{i}" for i in range(self.dims)]

    @property
    def dims(self):
        """Number of dimensions (parameters) in the samples."""
        if self.x is None:
            return 0
        return self.x.shape[1] if self.x.ndim > 1 else 1

    def to_numpy(self, dtype: Any | str | None = None):
        logger.debug("Converting samples to numpy arrays")
        import array_api_compat.numpy as np

        if dtype is not None:
            dtype = resolve_dtype(dtype, np)
        else:
            dtype = convert_dtype(self.dtype, np)
        return self.__class__(
            x=self.x,
            parameters=self.parameters,
            log_likelihood=self.log_likelihood,
            log_prior=self.log_prior,
            log_q=self.log_q,
            xp=np,
        )

    def to_namespace(self, xp, dtype: Any | str | None = None):
        if dtype is None:
            dtype = convert_dtype(self.dtype, xp)
        else:
            dtype = resolve_dtype(dtype, xp)
        logger.debug("Converting samples to {} namespace", xp)
        return self.__class__(
            x=self.x,
            parameters=self.parameters,
            log_likelihood=self.log_likelihood,
            log_prior=self.log_prior,
            log_q=self.log_q,
            xp=xp,
            device=self.device,
            dtype=dtype,
        )

    def array_to_namespace(self, x, dtype=None):
        """Convert an array to the same namespace as the samples"""
        kwargs = {}
        if dtype is not None:
            kwargs["dtype"] = resolve_dtype(dtype, self.xp)
        else:
            kwargs["dtype"] = self.dtype
        x = asarray(x, self.xp, **kwargs)
        if self.device:
            x = to_device(x, self.device)
        return x

    def to_dict(self, flat: bool = True):
        samples = dict(zip(self.parameters, self.x.T, strict=True))
        out = {
            "log_likelihood": self.log_likelihood,
            "log_prior": self.log_prior,
            "log_q": self.log_q,
        }
        if flat:
            out.update(samples)
        else:
            out["samples"] = samples
        return out

    def to_dataframe(self, flat: bool = True):
        import pandas as pd

        return pd.DataFrame(self.to_dict(flat=flat))

    def plot_corner(
        self,
        parameters: list[str] | None = None,
        fig: Figure | None = None,
        **kwargs,
    ):
        """Plot a corner plot of the samples.

        Parameters
        ----------
        parameters : list[str] | None
            List of parameters to plot. If None, all parameters are plotted.
        fig : matplotlib.figure.Figure | None
            Figure to plot on. If None, a new figure is created.
        **kwargs : dict
            Additional keyword arguments to pass to corner.corner(). Kwargs
            are deep-copied before use.
        """
        import corner

        kwargs = copy.deepcopy(kwargs)
        kwargs.setdefault("labels", self.parameters)

        if parameters is not None:
            indices = [self.parameters.index(p) for p in parameters]
            kwargs["labels"] = parameters
            x = self.x[:, indices] if self.x.ndim > 1 else self.x[indices]
        else:
            x = self.x
        fig = corner.corner(to_numpy(x), fig=fig, **kwargs)
        return fig

    def __str__(self):
        out = (
            f"No. samples: {len(self.x)}\nNo. parameters: {self.x.shape[-1]}\n"
        )
        return out

    def save(self, h5_file, path="samples", flat=False):
        """Save the samples to an HDF5 file.

        This converts the samples to numpy and then to a dictionary.

        Parameters
        ----------
        h5_file : h5py.File
            The HDF5 file to save to.
        path : str
            The path in the HDF5 file to save to.
        flat : bool
            If True, save the samples as a flat dictionary.
            If False, save the samples as a nested dictionary.
        """
        dictionary = self.to_numpy().to_dict(flat=flat)
        recursively_save_to_h5_file(h5_file, path, dictionary)

    def __len__(self):
        return len(self.x)

    def __getitem__(self, idx) -> BaseSamples:
        return self.__class__(
            x=self.x[idx],
            log_likelihood=self.log_likelihood[idx]
            if self.log_likelihood is not None
            else None,
            log_prior=self.log_prior[idx]
            if self.log_prior is not None
            else None,
            log_q=self.log_q[idx] if self.log_q is not None else None,
            parameters=self.parameters,
            dtype=self.dtype,
        )

    def __setitem__(self, idx, value: BaseSamples):
        raise NotImplementedError("Setting items is not supported")

    @classmethod
    def concatenate(cls, samples: list[BaseSamples]) -> BaseSamples:
        """Concatenate multiple Samples objects."""
        if not samples:
            raise ValueError("No samples to concatenate")
        if not all(s.parameters == samples[0].parameters for s in samples):
            raise ValueError("Parameters do not match")
        if not all(s.xp == samples[0].xp for s in samples):
            raise ValueError("Array namespaces do not match")
        if not all(s.dtype == samples[0].dtype for s in samples):
            raise ValueError("Dtypes do not match")
        xp = samples[0].xp
        return cls(
            x=xp.concatenate([s.x for s in samples], axis=0),
            log_likelihood=xp.concatenate(
                [s.log_likelihood for s in samples], axis=0
            )
            if all(s.log_likelihood is not None for s in samples)
            else None,
            log_prior=xp.concatenate([s.log_prior for s in samples], axis=0)
            if all(s.log_prior is not None for s in samples)
            else None,
            log_q=xp.concatenate([s.log_q for s in samples], axis=0)
            if all(s.log_q is not None for s in samples)
            else None,
            parameters=samples[0].parameters,
            dtype=samples[0].dtype,
        )

    @classmethod
    def from_samples(cls, samples: BaseSamples, **kwargs) -> BaseSamples:
        """Create a Samples object from a BaseSamples object."""
        xp = kwargs.pop("xp", samples.xp)
        device = kwargs.pop("device", samples.device)
        dtype = kwargs.pop("dtype", samples.dtype)
        if dtype is not None:
            dtype = resolve_dtype(dtype, xp)
        else:
            dtype = convert_dtype(samples.dtype, xp)
        return cls(
            x=samples.x,
            log_likelihood=samples.log_likelihood,
            log_prior=samples.log_prior,
            log_q=samples.log_q,
            parameters=samples.parameters,
            xp=xp,
            device=device,
            **kwargs,
        )

    def __getstate__(self):
        state = self.__dict__.copy()
        # replace xp (callable) with module name string
        if self.xp is not None:
            state["xp"] = (
                self.xp.__name__ if hasattr(self.xp, "__name__") else None
            )
        return state

    def __setstate__(self, state):
        # Restore xp by checking the namespace of x
        state["xp"] = array_namespace(state["x"])
        self.__dict__.update(state)


@dataclass
class Samples(BaseSamples):
    """Class for storing samples and corresponding weights.

    If :code:`xp` is not specified, all inputs will be converted to match
    the array type of :code:`x`.
    """

    log_w: Array = field(init=False)
    weights: Array = field(init=False)
    evidence: float = field(init=False)
    evidence_error: float = field(init=False)
    log_evidence: float | None = None
    log_evidence_error: float | None = None
    effective_sample_size: float = field(init=False)

    def __post_init__(self):
        super().__post_init__()

        if all(
            x is not None
            for x in [self.log_likelihood, self.log_prior, self.log_q]
        ):
            self.compute_weights()
        else:
            self.log_w = None
            self.weights = None
            self.evidence = None
            self.evidence_error = None
            self.effective_sample_size = None

    @property
    def efficiency(self):
        """Efficiency of the weighted samples.

        Defined as ESS / number of samples.
        """
        if self.log_w is None:
            raise RuntimeError("Samples do not contain weights!")
        return self.effective_sample_size / len(self.x)

    def compute_weights(self):
        """Compute the posterior weights."""
        self.log_w = self.log_likelihood + self.log_prior - self.log_q
        self.log_evidence = asarray(logsumexp(self.log_w), self.xp) - math.log(
            len(self.x)
        )
        self.weights = self.xp.exp(self.log_w)
        self.evidence = self.xp.exp(self.log_evidence)
        n = len(self.x)
        self.evidence_error = self.xp.sqrt(
            self.xp.sum((self.weights - self.evidence) ** 2) / (n * (n - 1))
        )
        self.log_evidence_error = self.xp.abs(
            self.evidence_error / self.evidence
        )
        log_w = self.log_w - self.xp.max(self.log_w)
        self.effective_sample_size = self.xp.exp(
            asarray(logsumexp(log_w) * 2 - logsumexp(log_w * 2), self.xp)
        )

    @property
    def scaled_weights(self):
        return self.xp.exp(self.log_w - self.xp.max(self.log_w))

    def rejection_sample(self, rng=None):
        if rng is None:
            rng = np.random.default_rng()
        log_u = asarray(
            np.log(rng.uniform(size=len(self.x))), self.xp, device=self.device
        )
        log_w = self.log_w - self.xp.max(self.log_w)
        accept = log_w > log_u
        return self.__class__(
            x=self.x[accept],
            log_likelihood=self.log_likelihood[accept],
            log_prior=self.log_prior[accept],
            dtype=self.dtype,
        )

    def to_dict(self, flat: bool = True):
        samples = dict(zip(self.parameters, self.x.T, strict=True))
        out = super().to_dict(flat=flat)
        other = {
            "log_w": self.log_w,
            "weights": self.weights,
            "evidence": self.evidence,
            "log_evidence": self.log_evidence,
            "evidence_error": self.evidence_error,
            "log_evidence_error": self.log_evidence_error,
            "effective_sample_size": self.effective_sample_size,
        }
        out.update(other)
        if flat:
            out.update(samples)
        else:
            out["samples"] = samples
        return out

    def plot_corner(self, include_weights: bool = True, **kwargs):
        kwargs = copy.deepcopy(kwargs)
        if (
            include_weights
            and self.weights is not None
            and "weights" not in kwargs
        ):
            kwargs["weights"] = to_numpy(self.scaled_weights)
        return super().plot_corner(**kwargs)

    def __str__(self):
        out = super().__str__()
        if self.log_evidence is not None:
            out += f"Log evidence: {self.log_evidence:.2f} +/- {self.log_evidence_error:.2f}\n"
        if self.log_w is not None:
            out += (
                f"Effective sample size: {self.effective_sample_size:.1f}\n"
                f"Efficiency: {self.efficiency:.2f}\n"
            )
        return out

    def to_namespace(self, xp):
        return self.__class__(
            x=asarray(self.x, xp),
            parameters=self.parameters,
            log_likelihood=asarray(self.log_likelihood, xp)
            if self.log_likelihood is not None
            else None,
            log_prior=asarray(self.log_prior, xp)
            if self.log_prior is not None
            else None,
            log_q=asarray(self.log_q, xp) if self.log_q is not None else None,
            log_evidence=asarray(self.log_evidence, xp)
            if self.log_evidence is not None
            else None,
            log_evidence_error=asarray(self.log_evidence_error, xp)
            if self.log_evidence_error is not None
            else None,
        )

    def to_numpy(self):
        return self.__class__(
            x=to_numpy(self.x),
            parameters=self.parameters,
            log_likelihood=to_numpy(self.log_likelihood)
            if self.log_likelihood is not None
            else None,
            log_prior=to_numpy(self.log_prior)
            if self.log_prior is not None
            else None,
            log_q=to_numpy(self.log_q) if self.log_q is not None else None,
            log_evidence=self.log_evidence
            if self.log_evidence is not None
            else None,
            log_evidence_error=self.log_evidence_error
            if self.log_evidence_error is not None
            else None,
        )

    def __getitem__(self, idx):
        sliced = super().__getitem__(idx)
        return self.__class__.from_samples(
            sliced,
            log_evidence=self.log_evidence,
            log_evidence_error=self.log_evidence_error,
        )


@dataclass
class SMCSamples(BaseSamples):
    beta: float | None = None
    """Temperature parameter for the current samples."""
    log_evidence: float | None = None
    """Log evidence estimate for the current samples."""
    log_evidence_error: float | None = None
    """Log evidence error estimate for the current samples."""

    def log_p_t(self, beta):
        log_p_T = self.log_likelihood + self.log_prior
        return (1 - beta) * self.log_q + beta * log_p_T

    def unnormalized_log_weights(self, beta: float) -> Array:
        return (self.beta - beta) * self.log_q + (beta - self.beta) * (
            self.log_likelihood + self.log_prior
        )

    def log_evidence_ratio(self, beta: float) -> float:
        log_w = self.unnormalized_log_weights(beta)
        return logsumexp(log_w) - math.log(len(self.x))

    def log_evidence_ratio_variance(self, beta: float) -> float:
        """Estimate the variance of the log evidence ratio using the delta method.

        Defined as Var(log Z) = Var(w) / (E[w])^2 where w are the unnormalized weights.
        """
        log_w = self.unnormalized_log_weights(beta)
        m = self.xp.max(log_w)
        u = self.xp.exp(log_w - m)
        mean_w = self.xp.mean(u)
        var_w = self.xp.var(u)
        return (
            var_w / (len(self) * (mean_w**2)) if mean_w != 0 else self.xp.nan
        )

    def log_weights(self, beta: float) -> Array:
        log_w = self.unnormalized_log_weights(beta)
        if self.xp.isnan(log_w).any():
            raise ValueError(f"Log weights contain NaN values for beta={beta}")
        log_evidence_ratio = logsumexp(log_w) - math.log(len(self.x))
        return log_w + log_evidence_ratio

    def resample(
        self,
        beta,
        n_samples: int | None = None,
        rng: np.random.Generator = None,
    ) -> "SMCSamples":
        if beta == self.beta and n_samples is None:
            logger.warning(
                "Resampling with the same beta value, returning identical samples"
            )
            return self
        if rng is None:
            rng = np.random.default_rng()
        if n_samples is None:
            n_samples = len(self.x)
        log_w = self.log_weights(beta)
        w = to_numpy(self.xp.exp(log_w - logsumexp(log_w)))
        idx = rng.choice(len(self.x), size=n_samples, replace=True, p=w)
        return self.__class__(
            x=self.x[idx],
            log_likelihood=self.log_likelihood[idx],
            log_prior=self.log_prior[idx],
            log_q=self.log_q[idx],
            beta=beta,
            dtype=self.dtype,
        )

    def __str__(self):
        out = super().__str__()
        if self.log_evidence is not None:
            out += f"Log evidence: {self.log_evidence:.2f}\n"
        return out

    def to_standard_samples(self):
        """Convert the samples to standard samples."""
        return Samples(
            x=self.x,
            log_likelihood=self.log_likelihood,
            log_prior=self.log_prior,
            xp=self.xp,
            parameters=self.parameters,
            log_evidence=self.log_evidence,
            log_evidence_error=self.log_evidence_error,
        )

    def __getitem__(self, idx):
        sliced = super().__getitem__(idx)
        return self.__class__.from_samples(
            sliced,
            beta=self.beta,
            log_evidence=self.log_evidence,
            log_evidence_error=self.log_evidence_error,
        )



import logging
from functools import partial

import numpy as np

from ...samples import SMCSamples
from ...utils import asarray, to_numpy, track_calls
from .base import SMCSampler

logger = logging.getLogger(__name__)


class BlackJAXSMC(SMCSampler):
    """BlackJAX SMC sampler."""

    def __init__(
        self,
        log_likelihood,
        log_prior,
        dims,
        prior_flow,
        xp,
        dtype=None,
        parameters=None,
        preconditioning_transform=None,
        rng: np.random.Generator | None = None,  # New parameter
    ):
        # For JAX compatibility, we'll keep the original xp
        super().__init__(
            log_likelihood=log_likelihood,
            log_prior=log_prior,
            dims=dims,
            prior_flow=prior_flow,
            xp=xp,
            dtype=dtype,
            parameters=parameters,
            preconditioning_transform=preconditioning_transform,
        )
        self.key = None
        self.rng = rng or np.random.default_rng()

    def log_prob(self, x, beta=None):
        """Log probability function compatible with BlackJAX."""
        # Convert to original xp format for computation
        if hasattr(x, "__array__"):
            x_original = asarray(x, self.xp)
        else:
            x_original = x

        # Transform back to parameter space
        x_params, log_abs_det_jacobian = (
            self.preconditioning_transform.inverse(x_original)
        )
        samples = SMCSamples(x_params, xp=self.xp, dtype=self.dtype)

        # Compute log probabilities
        log_q = self.prior_flow.log_prob(samples.x)
        samples.log_q = samples.array_to_namespace(log_q)
        samples.log_prior = samples.array_to_namespace(self.log_prior(samples))
        samples.log_likelihood = samples.array_to_namespace(
            self.log_likelihood(samples)
        )

        # Compute target log probability
        log_prob = samples.log_p_t(
            beta=beta
        ).flatten() + samples.array_to_namespace(log_abs_det_jacobian)

        # Handle NaN values
        log_prob = self.xp.where(
            self.xp.isnan(log_prob), -self.xp.inf, log_prob
        )

        return log_prob

    @track_calls
    def sample(
        self,
        n_samples: int,
        n_steps: int = None,
        adaptive: bool = True,
        target_efficiency: float = 0.5,
        target_efficiency_rate: float = 1.0,
        n_final_samples: int | None = None,
        sampler_kwargs: dict | None = None,
        rng_key=None,
    ):
        """Sample using BlackJAX SMC.

        Parameters
        ----------
        n_samples : int
            Number of samples to draw.
        n_steps : int
            Number of SMC steps.
        adaptive : bool
            Whether to use adaptive tempering.
        target_efficiency : float
            Target efficiency for adaptive tempering.
        n_final_samples : int | None
            Number of final samples to return.
        sampler_kwargs : dict | None
            Additional arguments for the BlackJAX sampler.
            - algorithm: str, one of "nuts", "hmc", "rwmh", "random_walk"
            - n_steps: int, number of MCMC steps per mutation
            - step_size: float, step size for HMC/NUTS
            - inverse_mass_matrix: array, inverse mass matrix
            - sigma: float or array, proposal covariance for random walk MH
            - num_integration_steps: int, integration steps for HMC
        rng_key : jax.random.key| None
            JAX random key for reproducibility.
        """
        self.sampler_kwargs = sampler_kwargs or {}
        self.sampler_kwargs.setdefault("n_steps", 5 * self.dims)
        self.sampler_kwargs.setdefault("algorithm", "nuts")
        self.sampler_kwargs.setdefault("step_size", 1e-3)
        self.sampler_kwargs.setdefault("inverse_mass_matrix", None)
        self.sampler_kwargs.setdefault("sigma", 0.1)  # For random walk MH

        # Initialize JAX random key
        if rng_key is None:
            import jax

            self.key = jax.random.key(42)
        else:
            self.key = rng_key

        return super().sample(
            n_samples,
            n_steps=n_steps,
            adaptive=adaptive,
            target_efficiency=target_efficiency,
            target_efficiency_rate=target_efficiency_rate,
            n_final_samples=n_final_samples,
        )

    def mutate(self, particles, beta, n_steps=None):
        """Mutate particles using BlackJAX MCMC."""
        import blackjax
        import jax

        logger.debug("Mutating particles with BlackJAX")

        # Split the random key
        self.key, subkey = jax.random.split(self.key)

        # Transform particles to latent space
        z = self.fit_preconditioning_transform(particles.x)

        # Convert to JAX arrays
        z_jax = jax.numpy.asarray(to_numpy(z))

        # Create log probability function for this beta
        log_prob_fn = partial(self._jax_log_prob, beta=beta)

        # Choose BlackJAX algorithm
        algorithm = self.sampler_kwargs["algorithm"].lower()

        n_steps = n_steps or self.sampler_kwargs["n_steps"]

        if algorithm == "rwmh" or algorithm == "random_walk":
            # Initialize Random Walk Metropolis-Hastings sampler
            sigma = self.sampler_kwargs.get("sigma", 0.1)

            # BlackJAX RMH expects a transition function, not a covariance
            if isinstance(sigma, (int, float)):
                # Create a multivariate normal proposal function
                def proposal_fn(key, position):
                    return position + sigma * jax.random.normal(
                        key, position.shape
                    )
            else:
                # For more complex covariance structures
                if len(sigma) == self.dims:
                    # Diagonal covariance
                    sigma_diag = jax.numpy.array(sigma)

                    def proposal_fn(key, position):
                        return position + sigma_diag * jax.random.normal(
                            key, position.shape
                        )
                else:
                    # Full covariance matrix
                    sigma_matrix = jax.numpy.array(sigma)

                    def proposal_fn(key, position):
                        return position + jax.random.multivariate_normal(
                            key, jax.numpy.zeros(self.dims), sigma_matrix
                        )

            rwmh = blackjax.rmh(log_prob_fn, proposal_fn)

            # Initialize states for each particle
            n_particles = z_jax.shape[0]
            keys = jax.random.split(subkey, n_particles)

            # Vectorized initialization and sampling
            def init_and_sample(key, z_init):
                state = rwmh.init(z_init)

                def one_step(state, key):
                    state, info = rwmh.step(key, state)
                    return state, (state, info)

                keys = jax.random.split(key, n_steps)
                final_state, (states, infos) = jax.lax.scan(
                    one_step, state, keys
                )
                return final_state, infos

            # Vectorize over particles
            final_states, all_infos = jax.vmap(init_and_sample)(keys, z_jax)

            # Extract final positions
            z_final = final_states.position

            # Calculate acceptance rates
            acceptance_rates = jax.numpy.mean(all_infos.is_accepted, axis=1)
            mean_acceptance = jax.numpy.mean(acceptance_rates)

        elif algorithm == "nuts":
            # Initialize step size and mass matrix if not provided
            inverse_mass_matrix = self.sampler_kwargs.get(
                "inverse_mass_matrix"
            )
            if inverse_mass_matrix is None:
                inverse_mass_matrix = jax.numpy.eye(self.dims)

            step_size = self.sampler_kwargs["step_size"]

            # Initialize NUTS sampler
            nuts = blackjax.nuts(
                log_prob_fn,
                step_size=step_size,
                inverse_mass_matrix=inverse_mass_matrix,
            )

            # Initialize states for each particle
            n_particles = z_jax.shape[0]
            keys = jax.random.split(subkey, n_particles)

            # Vectorized initialization and sampling
            def init_and_sample(key, z_init):
                state = nuts.init(z_init)

                def one_step(state, key):
                    state, info = nuts.step(key, state)
                    return state, (state, info)

                keys = jax.random.split(key, self.sampler_kwargs["n_steps"])
                final_state, (states, infos) = jax.lax.scan(
                    one_step, state, keys
                )
                return final_state, infos

            # Vectorize over particles
            final_states, all_infos = jax.vmap(init_and_sample)(keys, z_jax)

            # Extract final positions
            z_final = final_states.position

            # Calculate acceptance rates
            try:
                acceptance_rates = jax.numpy.mean(
                    all_infos.is_accepted, axis=1
                )
                mean_acceptance = jax.numpy.mean(acceptance_rates)
            except AttributeError:
                mean_acceptance = np.nan

        elif algorithm == "hmc":
            # Initialize HMC sampler
            hmc = blackjax.hmc(
                log_prob_fn,
                step_size=self.sampler_kwargs["step_size"],
                num_integration_steps=self.sampler_kwargs.get(
                    "num_integration_steps", 10
                ),
                inverse_mass_matrix=(
                    self.sampler_kwargs["inverse_mass_matrix"]
                    or jax.numpy.eye(self.dims)
                ),
            )

            # Similar vectorized sampling as NUTS
            n_particles = z_jax.shape[0]
            keys = jax.random.split(subkey, n_particles)

            def init_and_sample(key, z_init):
                state = hmc.init(z_init)

                def one_step(state, key):
                    state, info = hmc.step(key, state)
                    return state, (state, info)

                keys = jax.random.split(key, self.sampler_kwargs["n_steps"])
                final_state, (states, infos) = jax.lax.scan(
                    one_step, state, keys
                )
                return final_state, infos

            final_states, all_infos = jax.vmap(init_and_sample)(keys, z_jax)
            z_final = final_states.position
            try:
                acceptance_rates = jax.numpy.mean(
                    all_infos.is_accepted, axis=1
                )
                mean_acceptance = jax.numpy.mean(acceptance_rates)
            except AttributeError:
                mean_acceptance = np.nan

        else:
            raise ValueError(f"Unsupported algorithm: {algorithm}")

        # Convert back to parameter space
        z_final_np = to_numpy(z_final)
        x_final = self.preconditioning_transform.inverse(z_final_np)[0]

        # Store MCMC diagnostics
        self.history.mcmc_acceptance.append(float(mean_acceptance))

        # Create new samples
        samples = SMCSamples(x_final, xp=self.xp, beta=beta, dtype=self.dtype)
        samples.log_q = samples.array_to_namespace(
            self.prior_flow.log_prob(samples.x)
        )
        samples.log_prior = samples.array_to_namespace(self.log_prior(samples))
        samples.log_likelihood = samples.array_to_namespace(
            self.log_likelihood(samples)
        )

        if samples.xp.isnan(samples.log_q).any():
            raise ValueError("Log proposal contains NaN values")

        return samples

    def _jax_log_prob(self, z, beta):
        """JAX-compatible log probability function."""
        import jax.numpy as jnp

        # Single particle version for JAX
        z_expanded = jnp.expand_dims(z, 0)  # Add batch dimension
        log_prob = self.log_prob(z_expanded, beta=beta)
        return log_prob[0]  # Remove batch dimension



import math

import pytest

from aspire import transforms
from aspire.utils import AspireFile, copy_array, update_at_indices


def _make_array(xp, data, dtype):
    if dtype is None:
        return xp.asarray(data)
    return xp.asarray(data, dtype=dtype)


def save_and_load(tmp_path, transform):
    # Save the transform to an HDF5 file
    with AspireFile(tmp_path / "result.h5", "w") as h5_file:
        transform.save(h5_file, path="flow/data_transform")

    # Load the transform from the HDF5 file
    with AspireFile(tmp_path / "result.h5", "r") as h5_file:
        loaded_transform = transform.__class__.load(
            h5_file, path="flow/data_transform"
        )
    return loaded_transform


def test_save_and_load_identity_transform(tmp_path, xp, dtype):
    data_transform = transforms.IdentityTransform(xp=xp, dtype=dtype)
    loaded_transform = save_and_load(tmp_path, data_transform)

    # Check that the loaded transform has the same parameters
    assert type(loaded_transform) is type(data_transform)
    assert loaded_transform.dtype == data_transform.dtype


def test_identity_transform_forward_inverse_roundtrip(xp, dtype):
    transform = transforms.IdentityTransform(xp=xp, dtype=dtype)
    x = _make_array(xp, [[0.0, 1.0], [1.5, -2.5]], dtype)

    y, log_j = transform.forward(x)
    x_inv, inv_log_j = transform.inverse(y)

    assert xp.allclose(y, x)
    assert xp.allclose(x_inv, x)
    assert xp.allclose(log_j, xp.zeros(len(x)))
    assert xp.allclose(inv_log_j, xp.zeros(len(x)))


def test_save_and_load_periodic_transform(tmp_path, xp, dtype):
    data_transform = transforms.PeriodicTransform(
        lower=0, upper=xp.pi, xp=xp, dtype=dtype
    )
    loaded_transform = save_and_load(tmp_path, data_transform)

    # Check that the loaded transform has the same parameters
    assert type(loaded_transform) is type(data_transform)
    assert loaded_transform.dtype == data_transform.dtype


def test_periodic_transform_wraps_and_inverts(xp, dtype):
    lower = 0.0
    upper = 2 * math.pi
    transform = transforms.PeriodicTransform(
        lower=lower, upper=upper, xp=xp, dtype=dtype
    )
    width = upper - lower
    raw = [-math.pi, 3 * math.pi]
    x = _make_array(xp, [[raw[0]], [raw[1]]], dtype)
    expected = [((value - lower) % width) + lower for value in raw]
    expected_arr = _make_array(xp, [[v] for v in expected], transform.dtype)

    y, log_j = transform.forward(x)
    x_inv, inv_log_j = transform.inverse(y)

    assert xp.allclose(y, expected_arr, atol=1e-6)
    assert xp.allclose(x_inv, expected_arr, atol=1e-6)
    assert xp.allclose(log_j, xp.zeros(y.shape[0]))
    assert xp.allclose(inv_log_j, xp.zeros(y.shape[0]))


def test_periodic_transform_multidimensional(xp, dtype):
    lower = _make_array(xp, [0.0, -math.pi], dtype)
    upper = _make_array(xp, [2 * math.pi, math.pi], dtype)
    transform = transforms.PeriodicTransform(
        lower=lower, upper=upper, xp=xp, dtype=dtype
    )
    x = _make_array(
        xp,
        [
            [-3.0, -3.5],
            [7.0, 3.2],
        ],
        dtype,
    )

    y, log_j = transform.forward(x)
    x_inv, inv_log_j = transform.inverse(y)

    assert y.shape == x.shape
    assert xp.allclose(x_inv, y, atol=1e-6)
    assert log_j.shape == (x.shape[0],)
    assert inv_log_j.shape == (x.shape[0],)


def test_save_and_load_affine_transform(tmp_path, rng, xp):
    dims = 3

    x = xp.asarray(rng.normal(size=(100, dims)))
    data_transform = transforms.AffineTransform(xp)
    data_transform.fit(x)

    loaded_transform = save_and_load(tmp_path, data_transform)

    # Check that the loaded transform has the same parameters
    assert loaded_transform._mean.shape == data_transform._mean.shape
    assert loaded_transform._std.shape == data_transform._std.shape
    # Check types are the same
    assert type(loaded_transform._mean) is type(data_transform._mean)
    assert type(loaded_transform._std) is type(data_transform._std)
    # Check values are close
    assert xp.allclose(loaded_transform._mean, data_transform._mean)
    assert xp.allclose(loaded_transform._std, data_transform._std)


def test_logit_transform_forward_inverse_roundtrip(xp, dtype):
    transform = transforms.LogitTransform(
        lower=-1.0, upper=2.0, xp=xp, dtype=dtype
    )
    x = _make_array(
        xp,
        [
            [-0.75, 0.0, 1.25],
            [1.5, -0.25, 0.8],
        ],
        dtype,
    )

    y, log_j = transform.forward(x)
    x_inv, inv_log_j = transform.inverse(y)

    assert xp.allclose(x_inv, x, atol=1e-5)
    assert log_j.shape == (x.shape[0],)
    assert inv_log_j.shape == (x.shape[0],)


def test_logit_transform_multidimensional(xp, dtype):
    lower = _make_array(xp, [-1.0, 0.0], dtype)
    upper = _make_array(xp, [1.0, 2.0], dtype)
    transform = transforms.LogitTransform(
        lower=lower, upper=upper, xp=xp, dtype=dtype
    )
    x = _make_array(
        xp,
        [
            [-0.5, 0.5],
            [0.75, 1.75],
        ],
        dtype,
    )

    y, log_j = transform.forward(x)
    x_inv, inv_log_j = transform.inverse(y)

    assert xp.allclose(x_inv, x, atol=1e-5)
    assert log_j.shape == (x.shape[0],)
    assert inv_log_j.shape == (x.shape[0],)


def test_save_and_load_logit_transform(tmp_path, xp, dtype):
    data_transform = transforms.LogitTransform(
        lower=-2, upper=3, xp=xp, dtype=dtype
    )
    loaded_transform = save_and_load(tmp_path, data_transform)

    # Check that the loaded transform has the same parameters
    assert loaded_transform.lower == data_transform.lower
    assert loaded_transform.upper == data_transform.upper
    assert loaded_transform.eps == data_transform.eps
    assert loaded_transform.dtype == data_transform.dtype


def test_save_and_load_probit_transform(tmp_path, xp, dtype):
    data_transform = transforms.ProbitTransform(
        lower=-2, upper=3, xp=xp, dtype=dtype
    )
    loaded_transform = save_and_load(tmp_path, data_transform)

    # Check that the loaded transform has the same parameters
    assert loaded_transform.lower == data_transform.lower
    assert loaded_transform.upper == data_transform.upper
    assert loaded_transform.eps == data_transform.eps
    assert loaded_transform.dtype == data_transform.dtype


def test_save_and_load_composite_transform(tmp_path, rng, xp, dtype):
    dims = 3

    parameters = [f"x_{i}" for i in range(dims)]
    x = xp.asarray(rng.normal(size=(100, dims)))

    transform = transforms.CompositeTransform(
        parameters=parameters,
        periodic_parameters=["x_0"],
        prior_bounds={p: [-3, 3] for p in parameters},
        xp=xp,
        dtype=dtype,
    )
    transform.fit(x)

    loaded_transform = save_and_load(tmp_path, transform)

    # Check that the loaded transform has the same parameters
    assert type(loaded_transform) is type(transform)
    assert loaded_transform.dtype == transform.dtype
    assert loaded_transform.parameters == transform.parameters
    assert (
        loaded_transform.periodic_parameters == transform.periodic_parameters
    )

    x_forward, _ = transform.forward(x)
    x_inverse, _ = transform.inverse(x_forward)

    x_forward_loaded, _ = loaded_transform.forward(x)
    x_inverse_loaded, _ = loaded_transform.inverse(x_forward_loaded)

    # Check that the forward and inverse transforms are the same
    assert xp.allclose(x_forward, x_forward_loaded)
    assert xp.allclose(x_inverse, x_inverse_loaded)


def test_composite_transform_forward_inverse_roundtrip(xp, dtype):
    parameters = ["x0", "x1", "x2"]
    prior_bounds = {p: (-3.0, 3.0) for p in parameters}
    transform = transforms.CompositeTransform(
        parameters=parameters,
        periodic_parameters=["x0"],
        prior_bounds=prior_bounds,
        xp=xp,
        dtype=dtype,
        affine_transform=True,
    )

    assert transform._periodic_transform.dtype == transform.dtype
    assert transform._bounded_transform.dtype == transform.dtype
    assert transform._affine_transform.dtype == transform.dtype

    x_fit = _make_array(
        xp,
        [
            [-2.0, -2.5, 0.0],
            [2.0, 1.5, -1.0],
            [0.0, 0.0, 0.0],
        ],
        dtype,
    )
    transform.fit(x_fit)

    x = _make_array(
        xp,
        [
            [-4.0, -2.5, 0.0],
            [7.0, 1.5, -1.0],
            [0.0, 0.0, 0.0],
        ],
        dtype,
    )

    y, log_j = transform.forward(x)
    x_inv, inv_log_j = transform.inverse(y)

    x_exp = copy_array(x)
    print(x_exp)
    print((x_exp[:, 0] + 3) % 6 - 3)
    x_exp = update_at_indices(
        x_exp,
        (slice(None), 0),
        ((x_exp[:, 0] + 3) % 6) - 3,
    )  # Wrap x0 to [-3, 3]
    print(x_exp)

    assert x.shape == y.shape
    assert xp.allclose(x_inv, x_exp, atol=1e-5)
    assert log_j.shape == (x.shape[0],)
    assert inv_log_j.shape == (x.shape[0],)


def test_composite_transform_requires_prior_bounds_for_periodic(xp, dtype):
    with pytest.raises(ValueError):
        transforms.CompositeTransform(
            parameters=[0],
            periodic_parameters=[0],
            prior_bounds=None,
            xp=xp,
            dtype=dtype,
        )


def test_composite_transform_new_instance_copies_configuration(xp, dtype):
    parameters = ["x0", "x1"]
    prior_bounds = {p: (-1.0, 1.0) for p in parameters}
    transform = transforms.CompositeTransform(
        parameters=parameters,
        periodic_parameters=[],
        prior_bounds=prior_bounds,
        xp=xp,
        dtype=dtype,
        affine_transform=False,
    )

    new_transform = transform.new_instance()
    assert new_transform.xp is transform.xp
    assert new_transform.dtype == transform.dtype
    assert new_transform.parameters == transform.parameters
    assert new_transform.periodic_parameters == transform.periodic_parameters


def test_flow_transform_config_drops_periodic_parameters(xp, dtype):
    parameters = ["a", "b"]
    prior_bounds = {p: (-2.0, 2.0) for p in parameters}
    transform = transforms.FlowTransform(
        parameters=parameters,
        prior_bounds=prior_bounds,
        xp=xp,
        dtype=dtype,
    )

    cfg = transform.config_dict()
    assert "periodic_parameters" not in cfg
    assert cfg["parameters"] == parameters


def test_logit_transform_interval_check_raises(xp):
    with pytest.raises(ValueError):
        transforms.LogitTransform(lower=0.0, upper=0.0, xp=xp, dtype=None)


def test_probit_transform_multidimensional_roundtrip(xp, dtype):
    lower = _make_array(xp, [-2.0, -1.0], dtype)
    upper = _make_array(xp, [2.0, 3.0], dtype)
    transform = transforms.ProbitTransform(
        lower=lower, upper=upper, xp=xp, dtype=dtype
    )
    x = _make_array(
        xp,
        [
            [-1.0, -0.5],
            [1.0, 2.5],
        ],
        dtype,
    )

    y, log_j = transform.forward(x)
    x_inv, inv_log_j = transform.inverse(y)

    assert xp.allclose(x_inv, x, atol=1e-5)
    assert log_j.shape == (x.shape[0],)
    assert inv_log_j.shape == (x.shape[0],)



import copy


def plot_comparison(
    *samples, parameters=None, per_samples_kwargs=None, labels=None, **kwargs
):
    """
    Plot a comparison of multiple samples.
    """
    default_kwargs = dict(
        density=True,
        bins=30,
        color="C0",
        smooth=1.0,
        plot_datapoints=True,
        plot_density=False,
        hist_kwargs=dict(density=True, color="C0"),
    )
    default_kwargs.update(kwargs)

    if per_samples_kwargs is None:
        per_samples_kwargs = [{}] * len(samples)

    fig = None
    for i, sample in enumerate(samples):
        kwds = copy.deepcopy(default_kwargs)
        color = per_samples_kwargs[i].pop("color", f"C{i}")
        kwds["color"] = color
        kwds["hist_kwargs"]["color"] = color
        kwds.update(per_samples_kwargs[i])
        fig = sample.plot_corner(fig=fig, parameters=parameters, **kwds)

    if labels:
        fig.legend(
            labels=labels,
            loc="upper right",
            bbox_to_anchor=(0.9, 0.9),
            bbox_transform=fig.transFigure,
        )
    return fig


def plot_history_comparison(*histories):
    # Assert that all histories are of the same type
    if not all(isinstance(h, histories[0].__class__) for h in histories):
        raise ValueError("All histories must be of the same type")
    fig = histories[0].plot()
    for history in histories[1:]:
        fig = history.plot(fig=fig)
    return fig



import importlib
import logging
import math
from typing import Any, Callable

import h5py
from array_api_compat import device as get_device
from array_api_compat import is_torch_namespace
from array_api_compat.common._typing import Array

from .flows import get_flow_wrapper
from .utils import (
    asarray,
    convert_dtype,
    copy_array,
    logit,
    sigmoid,
    update_at_indices,
)

logger = logging.getLogger(__name__)


class BaseTransform:
    """Base class for data transforms.

    Parameters
    ----------
    xp : Callable
        The array API namespace to use (e.g., numpy, torch).
    dtype : Any, optional
        The data type to use for the transform. If not provided, defaults to
        the default dtype of the array API namespace if available.
    """

    def __init__(self, xp, dtype=None):
        self.xp = xp
        if is_torch_namespace(self.xp) and dtype is None:
            dtype = self.xp.get_default_dtype()
        elif isinstance(dtype, str):
            from .utils import resolve_dtype

            dtype = resolve_dtype(dtype, self.xp)
        self.dtype = dtype

    def fit(self, x):
        """Fit the transform to the data."""
        raise NotImplementedError("Subclasses must implement fit method.")

    def forward(self, x):
        raise NotImplementedError("Subclasses must implement forward method.")

    def inverse(self, y):
        raise NotImplementedError("Subclasses must implement inverse method.")

    def config_dict(self):
        """Return the configuration of the transform as a dictionary."""
        return {
            "xp": self.xp.__name__,
            "dtype": str(self.dtype) if self.dtype else None,
        }

    def save(self, h5_file: h5py.File, path: str = "data_transform"):
        """Save config + any fitted state into an HDF5 file."""
        from .utils import encode_dtype, recursively_save_to_h5_file

        # store class name for reconstruction
        grp = h5_file.create_group(path)
        grp.attrs["class"] = self.__class__.__name__
        # store config as JSON
        config = self.config_dict()
        config["dtype"] = encode_dtype(self.xp, config["dtype"])
        recursively_save_to_h5_file(grp, "config", config)
        # store any fitted arrays
        self._save_state(grp)

    @classmethod
    def load(
        cls,
        h5_file: h5py.File,
        path: str = "data_transform",
        strict: bool = False,
    ):
        """Reconstruct transform from file.

        Parameters
        ----------
        h5_file : h5py.File
            The HDF5 file to load from.
        path : str, optional
            The path in the HDF5 file where the transform is stored.
        strict : bool, optional
            If True, raise an error if the class in the file does not match cls.
            If False, load the class specified in the file. Default is False.
        """
        from .utils import decode_dtype, load_from_h5_file

        grp = h5_file[path]
        class_name = grp.attrs["class"]
        if class_name != cls.__name__:
            if strict:
                raise ValueError(
                    f"Expected class {cls.__name__}, got {class_name}."
                )
            else:
                cls = getattr(importlib.import_module(__name__), class_name)
                logger.info(
                    f"Loading class {class_name} instead of {cls.__name__}."
                )

        config = load_from_h5_file(grp, "config")
        config["xp"] = importlib.import_module(config["xp"])
        config["dtype"] = decode_dtype(config["xp"], config["dtype"])
        obj = cls(**config)
        obj._load_state(grp)
        return obj

    def _save_state(self, h5_file: h5py.File):
        pass

    def _load_state(self, h5_file: h5py.File):
        pass


class IdentityTransform(BaseTransform):
    """Identity transform that does nothing to the data."""

    def fit(self, x):
        return copy_array(x, xp=self.xp)

    def forward(self, x):
        return copy_array(x, xp=self.xp), self.xp.zeros(
            len(x), device=get_device(x)
        )

    def inverse(self, y):
        return copy_array(y, xp=self.xp), self.xp.zeros(
            len(y), device=get_device(y)
        )


class CompositeTransform(BaseTransform):
    def __init__(
        self,
        parameters: list[int],
        periodic_parameters: list[int] = None,
        prior_bounds: list[tuple[float, float]] = None,
        bounded_to_unbounded: bool = True,
        bounded_transform: str = "probit",
        affine_transform: bool = True,
        device=None,
        xp: None = None,
        eps: float = 1e-6,
        dtype: Any = None,
    ):
        super().__init__(xp=xp, dtype=dtype)
        if prior_bounds is None:
            logger.warning(
                "Missing prior bounds, some transforms may not be applied."
            )
        if periodic_parameters and not prior_bounds:
            raise ValueError(
                "Must specify prior bounds to use periodic parameters."
            )
        self.parameters = parameters
        self.periodic_parameters = periodic_parameters or []
        self.bounded_to_unbounded = bounded_to_unbounded
        self.bounded_transform = bounded_transform
        self.affine_transform = affine_transform

        self.eps = eps
        self.device = device

        if prior_bounds is None:
            self.prior_bounds = None
            self.bounded_parameters = None
            lower_bounds = None
            upper_bounds = None
        else:
            logger.info(f"Prior bounds: {prior_bounds}")
            self.prior_bounds = {
                k: self.xp.asarray(
                    prior_bounds[k], device=device, dtype=self.dtype
                )
                for k in self.parameters
            }
            if bounded_to_unbounded:
                self.bounded_parameters = [
                    p
                    for p in parameters
                    if self.xp.isfinite(self.prior_bounds[p]).all()
                    and p not in self.periodic_parameters
                ]
            else:
                self.bounded_parameters = None
            lower_bounds = self.xp.asarray(
                [self.prior_bounds[p][0] for p in parameters],
                device=device,
                dtype=self.dtype,
            )
            upper_bounds = self.xp.asarray(
                [self.prior_bounds[p][1] for p in parameters],
                device=device,
                dtype=self.dtype,
            )

        if self.periodic_parameters:
            logger.info(f"Periodic parameters: {self.periodic_parameters}")
            self.periodic_mask = self.xp.asarray(
                [p in self.periodic_parameters for p in parameters],
                dtype=bool,
                device=device,
            )
            self._periodic_transform = PeriodicTransform(
                lower=lower_bounds[self.periodic_mask],
                upper=upper_bounds[self.periodic_mask],
                xp=self.xp,
                dtype=self.dtype,
            )
        if self.bounded_parameters:
            logger.info(f"Bounded parameters: {self.bounded_parameters}")
            self.bounded_mask = self.xp.asarray(
                [p in self.bounded_parameters for p in parameters], dtype=bool
            )
            if self.bounded_transform == "probit":
                BoundedClass = ProbitTransform
            elif self.bounded_transform == "logit":
                BoundedClass = LogitTransform
            else:
                raise ValueError(
                    f"Unknown bounded transform: {self.bounded_transform}"
                )

            self._bounded_transform = BoundedClass(
                lower=lower_bounds[self.bounded_mask],
                upper=upper_bounds[self.bounded_mask],
                xp=self.xp,
                eps=self.eps,
                dtype=self.dtype,
            )

        if self.affine_transform:
            logger.info(f"Affine transform applied to: {self.parameters}")
            self._affine_transform = AffineTransform(
                xp=self.xp, dtype=self.dtype
            )
        else:
            self._affine_transform = None

    def fit(self, x):
        x = copy_array(x, xp=self.xp)
        if self.periodic_parameters:
            logger.debug(
                f"Fitting periodic transform to parameters: {self.periodic_parameters}"
            )
            x = update_at_indices(
                x,
                (slice(None), self.periodic_mask),
                self._periodic_transform.fit(x[:, self.periodic_mask]),
            )
        if self.bounded_parameters:
            logger.debug(
                f"Fitting bounded transform to parameters: {self.bounded_parameters}"
            )
            x = update_at_indices(
                x,
                (slice(None), self.bounded_mask),
                self._bounded_transform.fit(x[:, self.bounded_mask]),
            )
        if self.affine_transform:
            logger.debug("Fitting affine transform to all parameters.")
            x = self._affine_transform.fit(x)
        return x

    def forward(self, x):
        x = copy_array(x, xp=self.xp)
        x = self.xp.atleast_2d(x)
        log_abs_det_jacobian = self.xp.zeros(len(x), device=self.device)
        if self.periodic_parameters:
            y, log_j_periodic = self._periodic_transform.forward(
                x[..., self.periodic_mask]
            )
            x = update_at_indices(x, (slice(None), self.periodic_mask), y)
            log_abs_det_jacobian += log_j_periodic

        if self.bounded_parameters:
            y, log_j_bounded = self._bounded_transform.forward(
                x[..., self.bounded_mask]
            )
            x = update_at_indices(x, (slice(None), self.bounded_mask), y)
            log_abs_det_jacobian += log_j_bounded

        if self.affine_transform:
            x, log_j_affine = self._affine_transform.forward(x)
            log_abs_det_jacobian += log_j_affine
        return x, log_abs_det_jacobian

    def inverse(self, x):
        x = copy_array(x, xp=self.xp)
        x = self.xp.atleast_2d(x)
        log_abs_det_jacobian = self.xp.zeros(len(x), device=self.device)
        if self.affine_transform:
            x, log_j_affine = self._affine_transform.inverse(x)
            log_abs_det_jacobian += log_j_affine

        if self.bounded_parameters:
            y, log_j_bounded = self._bounded_transform.inverse(
                x[..., self.bounded_mask]
            )
            x = update_at_indices(x, (slice(None), self.bounded_mask), y)
            log_abs_det_jacobian += log_j_bounded

        if self.periodic_parameters:
            y, log_j_periodic = self._periodic_transform.inverse(
                x[..., self.periodic_mask]
            )
            x = update_at_indices(x, (slice(None), self.periodic_mask), y)
            log_abs_det_jacobian += log_j_periodic

        return x, log_abs_det_jacobian

    def new_instance(self, xp=None, dtype: Any = None):
        if xp is None:
            xp = self.xp
        if dtype is None:
            dtype = self.dtype
        dtype = convert_dtype(dtype, xp)

        return self.__class__(
            parameters=self.parameters,
            periodic_parameters=self.periodic_parameters,
            prior_bounds=self.prior_bounds,
            bounded_to_unbounded=self.bounded_to_unbounded,
            bounded_transform=self.bounded_transform,
            affine_transform=self.affine_transform,
            device=self.device,
            xp=xp or self.xp,
            eps=self.eps,
            dtype=dtype,
        )

    def _save_state(self, h5_file):
        if self.affine_transform:
            affine_grp = h5_file.create_group("affine_transform")
            self._affine_transform._save_state(affine_grp)

    def _load_state(self, h5_file):
        if self.affine_transform:
            affine_grp = h5_file["affine_transform"]
            self._affine_transform._load_state(affine_grp)

    def config_dict(self):
        return super().config_dict() | {
            "parameters": self.parameters,
            "periodic_parameters": self.periodic_parameters,
            "prior_bounds": self.prior_bounds,
            "bounded_to_unbounded": self.bounded_to_unbounded,
            "bounded_transform": self.bounded_transform,
            "affine_transform": self.affine_transform,
            "eps": self.eps,
            "device": self.device,
        }


class FlowTransform(CompositeTransform):
    """Subclass of CompositeTransform that uses a Flow for transformations.

    Does not support periodic transforms.
    """

    def __init__(
        self,
        parameters: list[int],
        prior_bounds: list[tuple[float, float]] = None,
        bounded_to_unbounded: bool = True,
        bounded_transform: str = "probit",
        affine_transform: bool = True,
        device=None,
        xp=None,
        eps=1e-6,
        dtype=None,
    ):
        super().__init__(
            parameters=parameters,
            periodic_parameters=[],
            prior_bounds=prior_bounds,
            bounded_to_unbounded=bounded_to_unbounded,
            bounded_transform=bounded_transform,
            affine_transform=affine_transform,
            device=device,
            xp=xp,
            eps=eps,
            dtype=dtype,
        )

    def new_instance(self, xp=None):
        return self.__class__(
            parameters=self.parameters,
            prior_bounds=self.prior_bounds,
            bounded_to_unbounded=self.bounded_to_unbounded,
            bounded_transform=self.bounded_transform,
            device=self.device,
            xp=xp or self.xp,
            eps=self.eps,
        )

    def config_dict(self):
        cfg = super().config_dict()
        cfg.pop(
            "periodic_parameters", None
        )  # Remove periodic_parameters from config
        return cfg


class PeriodicTransform(BaseTransform):
    name: str = "periodic"
    requires_prior_bounds: bool = True

    def __init__(self, lower, upper, xp, dtype=None):
        super().__init__(xp=xp, dtype=dtype)
        self.lower = xp.asarray(lower, dtype=self.dtype)
        self.upper = xp.asarray(upper, dtype=self.dtype)
        self._width = self.upper - self.lower
        self._shift = None

    def fit(self, x):
        return self.forward(x)[0]

    def forward(self, x):
        y = self.lower + (x - self.lower) % self._width
        return y, self.xp.zeros(y.shape[0], device=get_device(y))

    def inverse(self, y):
        x = self.lower + (y - self.lower) % self._width
        return x, self.xp.zeros(x.shape[0], device=get_device(x))

    def config_dict(self):
        return super().config_dict() | {
            "lower": self.lower.tolist(),
            "upper": self.upper.tolist(),
        }


class BoundedTransform(BaseTransform):
    """Base class for bounded transforms.

    Maps from [lower, upper] to [0, 1] and vice versa using a linear scaling.
    If the interval [lower, upper] is too small, it will shift by the midpoint.

    Must be subclassed to implement specific transforms (e.g., Probit, Logit).

    Parameters
    ----------
    lower : Array
        The lower bound of the interval.
    upper : Array
        The upper bound of the interval.
    xp : Callable
        The array API namespace to use (e.g., numpy, torch).
    dtype : Any, optional
        The data type to use for the transform. If not provided, defaults to
        the default dtype of the array API namespace if available.
    """

    name: str = "bounded"
    requires_prior_bounds: bool = True

    def __init__(
        self, lower: Array, upper: Array, xp: Callable, dtype: Any = None
    ):
        super().__init__(xp=xp, dtype=dtype)
        self.lower = xp.atleast_1d(xp.asarray(lower, dtype=self.dtype))
        self.upper = xp.atleast_1d(xp.asarray(upper, dtype=self.dtype))

        self.interval_check(self.lower, self.upper)

        self._denom = self.upper - self.lower
        self._scale_log_abs_det_jacobian = -xp.log(self._denom).sum()

    def to_unit_interval(self, x: Array) -> tuple[Array, Array]:
        """Map from [lower, upper] to [0, 1].

        Parameters
        ----------
        x : Array
            The input array to be mapped.

        Returns
        -------
        tuple[Array, Array]
            A tuple containing the mapped array and the log absolute determinant Jacobian.
        """
        y = (x - self.lower) / self._denom
        log_j = self._scale_log_abs_det_jacobian * self.xp.ones(
            y.shape[0], device=get_device(y)
        )
        return y, log_j

    def from_unit_interval(self, y: Array) -> tuple[Array, Array]:
        """Map from [0, 1] to [lower, upper].

        Parameters
        ----------
        y : Array
            The input array to be mapped.

        Returns
        -------
        tuple[Array, Array]
            A tuple containing the mapped array and the log absolute determinant Jacobian.
        """
        x = self._denom * y + self.lower
        log_j = -self._scale_log_abs_det_jacobian * self.xp.ones(
            x.shape[0], device=get_device(x)
        )
        return x, log_j

    def interval_check(self, lower: Array, upper: Array) -> bool:
        """Check if the interval [lower, upper] is too small"""
        if any((upper - lower) == 0.0):
            raise ValueError(
                f"Current floating precision ({self.dtype}) is too small for specified parameter ranges"
            )

    def fit(self, x):
        return self.forward(x)[0]

    def forward(self, x):
        raise NotImplementedError("Subclasses must implement forward method.")

    def inverse(self, y):
        raise NotImplementedError("Subclasses must implement inverse method.")

    def config_dict(self):
        return super().config_dict() | {
            "lower": self.lower.tolist(),
            "upper": self.upper.tolist(),
        }


class ProbitTransform(BoundedTransform):
    name: str = "probit"
    requires_prior_bounds: bool = True

    def __init__(self, lower, upper, xp, eps=1e-6, dtype=None):
        super().__init__(xp=xp, dtype=dtype, lower=lower, upper=upper)
        self.eps = eps

    def fit(self, x: Array) -> Array:
        return self.forward(x)[0]

    def forward(self, x: Array) -> tuple[Array, Array]:
        from scipy.special import erfinv

        y, log_j_unit = self.to_unit_interval(x)
        y = self.xp.clip(y, self.eps, 1.0 - self.eps)
        y = erfinv(2 * y - 1) * math.sqrt(2)
        log_abs_det_jacobian = 0.5 * (math.log(2 * math.pi) + y**2).sum(-1)
        log_abs_det_jacobian = log_abs_det_jacobian + log_j_unit
        return y, log_abs_det_jacobian

    def inverse(self, y: Array) -> tuple[Array, Array]:
        from scipy.special import erf

        log_abs_det_jacobian = -(0.5 * (math.log(2 * math.pi) + y**2)).sum(-1)
        x = 0.5 * (1 + erf(y / math.sqrt(2)))
        x, log_j_unit = self.from_unit_interval(x)
        log_abs_det_jacobian = log_abs_det_jacobian + log_j_unit
        return x, log_abs_det_jacobian

    def config_dict(self):
        return super().config_dict() | {
            "eps": self.eps,
        }


class LogitTransform(BoundedTransform):
    name: str = "logit"
    requires_prior_bounds: bool = True

    def __init__(
        self,
        lower: Array,
        upper: Array,
        xp: Callable,
        eps: float = 1e-6,
        dtype: Any = None,
    ):
        super().__init__(xp=xp, dtype=dtype, lower=lower, upper=upper)
        self.eps = eps

    def fit(self, x: Array) -> Array:
        return self.forward(x)[0]

    def forward(self, x: Array) -> tuple[Array, Array]:
        y, log_j_unit = self.to_unit_interval(x)
        y, log_abs_det_jacobian = logit(y, eps=self.eps)
        log_abs_det_jacobian = log_abs_det_jacobian + log_j_unit
        return y, log_abs_det_jacobian

    def inverse(self, y: Array) -> tuple[Array, Array]:
        x, log_abs_det_jacobian = sigmoid(y)
        x, log_j_unit = self.from_unit_interval(x)
        log_abs_det_jacobian = log_abs_det_jacobian + log_j_unit
        return x, log_abs_det_jacobian

    def config_dict(self) -> dict[str, Any]:
        return super().config_dict() | {
            "eps": self.eps,
        }


class AffineTransform(BaseTransform):
    name: str = "affine"
    requires_prior_bounds: bool = False

    def __init__(self, xp, dtype=None):
        super().__init__(xp=xp, dtype=dtype)
        self._mean = None
        self._std = None

    def fit(self, x):
        self._mean = x.mean(0)
        self._std = x.std(0)
        self.log_abs_det_jacobian = -self.xp.log(self.xp.abs(self._std)).sum()
        return self.forward(x)[0]

    def forward(self, x):
        y = (x - self._mean) / self._std
        return y, self.log_abs_det_jacobian * self.xp.ones(
            y.shape[0], device=get_device(y)
        )

    def inverse(self, y):
        x = y * self._std + self._mean
        return x, -self.log_abs_det_jacobian * self.xp.ones(
            y.shape[0], device=get_device(y)
        )

    def config_dict(self):
        return super().config_dict()

    def _save_state(self, h5_file):
        h5_file.create_dataset("mean", data=self._mean)
        h5_file.create_dataset("std", data=self._std)

    def _load_state(self, h5_file):
        self._mean = asarray(h5_file["mean"][()], xp=self.xp)
        self._std = asarray(h5_file["std"][()], xp=self.xp)
        self.log_abs_det_jacobian = -self.xp.log(self.xp.abs(self._std)).sum()


class FlowPreconditioningTransform(BaseTransform):
    def __init__(
        self,
        parameters: list[int],
        flow_backend: str = "zuko",
        prior_bounds: list[tuple[float, float]] = None,
        bounded_to_unbounded: bool = True,
        bounded_transform: str = "probit",
        affine_transform: bool = True,
        periodic_parameters: list[int] = None,
        device=None,
        xp=None,
        eps=1e-6,
        dtype=None,
        flow_matching: bool = False,
        flow_kwargs: dict[str, Any] = None,
        fit_kwargs: dict[str, Any] = None,
    ):
        super().__init__(xp=xp, dtype=dtype)

        self.parameters = parameters
        self.periodic_parameters = periodic_parameters or []
        self.prior_bounds = prior_bounds
        self.bounded_to_unbounded = bounded_to_unbounded
        self.bounded_transform = bounded_transform
        self.affine_transform = affine_transform
        self.eps = eps
        self.device = device or "cpu"
        self.flow_backend = flow_backend
        self.flow_matching = flow_matching
        self.flow_kwargs = dict(flow_kwargs or {})
        if dtype is not None:
            self.flow_kwargs.setdefault("dtype", dtype)
        self.fit_kwargs = dict(fit_kwargs or {})

        FlowClass, xp = get_flow_wrapper(
            backend=flow_backend, flow_matching=flow_matching
        )
        transform = CompositeTransform(
            parameters=parameters,
            periodic_parameters=periodic_parameters,
            prior_bounds=prior_bounds,
            bounded_to_unbounded=bounded_to_unbounded,
            bounded_transform=bounded_transform,
            affine_transform=affine_transform,
            device=device,
            xp=xp,
            eps=eps,
            dtype=dtype,
        )

        self._data_transform = transform
        self._FlowClass = FlowClass
        self.flow = None

    def fit(self, x):
        self.flow = self._FlowClass(
            dims=len(self.parameters),
            device=self.device,
            data_transform=self._data_transform,
            **self.flow_kwargs,
        )
        self.flow.fit(x, **self.fit_kwargs)
        return self.flow.forward(x, xp=self.xp)[0]

    def forward(self, x):
        return self.flow.forward(x, xp=self.xp)

    def inverse(self, y):
        return self.flow.inverse(y, xp=self.xp)

    def new_instance(self, xp=None, dtype: Any = None):
        if xp is None:
            xp = self.xp
        if dtype is None:
            dtype = self.dtype

        dtype = convert_dtype(dtype, xp)

        return self.__class__(
            parameters=self.parameters,
            periodic_parameters=self.periodic_parameters,
            prior_bounds=self.prior_bounds,
            bounded_to_unbounded=self.bounded_to_unbounded,
            bounded_transform=self.bounded_transform,
            affine_transform=self.affine_transform,
            device=self.device,
            xp=xp,
            eps=self.eps,
            dtype=dtype,
            flow_backend=self.flow_backend,
            flow_matching=self.flow_matching,
            flow_kwargs=self.flow_kwargs,
            fit_kwargs=self.fit_kwargs,
        )

    def save(self, h5_file, path="data_transform"):
        raise NotImplementedError(
            "FlowPreconditioningTransform does not support save method yet."
        )



"""
This examples demonstrates how to use aspire to fit a flow to a simple Gaussian
likelihood with a uniform prior.
"""

import math
from pathlib import Path

from scipy.stats import norm, uniform

from aspire import Aspire
from aspire.plot import plot_comparison
from aspire.samples import Samples
from aspire.utils import AspireFile, configure_logger

# Configure the logger
configure_logger("INFO")

outdir = Path("outdir") / "basic_example"
outdir.mkdir(parents=True, exist_ok=True)

# Number of dimensions
dims = 4


# Define the log likelihood and log prior
def log_likelihood(samples: Samples):
    # The log likelihood must accept a Samples object
    # The samples object contains the samples in the attribute samples.x
    return norm(2, 1).logpdf(samples.x).sum(axis=-1)


def log_prior(samples: Samples):
    return uniform(-10, 20).logpdf(samples.x).sum(axis=-1)


# True evidence is analytic for a Gaussian likelihood and uniform prior
true_log_evidence = -dims * math.log(20)

# Generate some initial samples
# These are slightly biased compared to the true posterior
initial_samples = Samples(norm(2.5, 1.0).rvs(size=(5000, dims)))
# Define the parameters and prior bounds
parameters = [f"x_{i}" for i in range(dims)]
prior_bounds = {p: [-10, 10] for p in parameters}

# Define the aspire object
aspire = Aspire(
    log_likelihood=log_likelihood,
    log_prior=log_prior,
    dims=dims,
    parameters=parameters,
    prior_bounds=prior_bounds,
)

# Fit the flow to the initial samples
history = aspire.fit(
    initial_samples,
    n_epochs=50,
)
# Plot the loss
fig = history.plot_loss()
fig.savefig(outdir / "loss.png")

# Produce samples from the posterior
samples = aspire.sample_posterior(5000)

# Save the the results to a file
# The AspireFile is a small wrapper around h5py.File that automatically
# includes additional metadata
with AspireFile(outdir / "aspire_result.h5", "w") as f:
    aspire.save_config(f, "aspire_config")
    samples.save(f, "posterior_samples")
    history.save(f, "flow_history")
    # Save the flow
    aspire.save_flow(f, "flow")

fig = plot_comparison(
    initial_samples,
    samples,
    samples,
    per_samples_kwargs=[
        dict(include_weights=True, color="C0"),
        dict(include_weights=False, color="lightgrey"),
        dict(include_weights=True, color="C1"),
    ],
    labels=["Training samples", "Samples (w/o weights)", "Posterior samples"],
)

fig.savefig(outdir / "comparison.png")
